#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for creation commands."""

from abc import abstractmethod
from enum import Enum as BuiltinEnum
from struct import calcsize, pack, unpack_from
from typing import Any, Dict, List, Mapping, Optional, Tuple, Type, Union

from spsdk import SPSDKError
from spsdk.sbfile.sb31.constants import EnumCmdTag
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import align_block, load_binary, value_to_int

########################################################################################################################
# Main Class
########################################################################################################################


class MainCmd:
    """Functions for creating cmd intended for inheritance."""

    def __eq__(self, obj: object) -> bool:
        """Comparison of values."""
        return isinstance(obj, self.__class__) and vars(obj) == vars(self)

    def __str__(self) -> str:
        """Get info of command."""
        return self.info()

    @abstractmethod
    def info(self) -> str:
        """Get info of command.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def export(self) -> bytes:
        """Export command as bytes.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> object:
        """Parse command from bytes array.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "MainCmd":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")


########################################################################################################################
# Base Command Class
########################################################################################################################


class BaseCmd(MainCmd):
    """Functions for creating cmd intended for inheritance."""

    FORMAT = "<4L"
    SIZE = calcsize(FORMAT)
    TAG = 0x55AAAA55

    @property
    def address(self) -> int:
        """Get address."""
        return self._address

    @address.setter
    def address(self, value: int) -> None:
        """Set address."""
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Invalid address")
        self._address = value

    @property
    def length(self) -> int:
        """Get length."""
        return self._length

    @length.setter
    def length(self, value: int) -> None:
        """Set value."""
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Invalid length")
        self._length = value

    def __init__(self, address: int, length: int, cmd_tag: int = EnumCmdTag.NONE) -> None:
        """Constructor for Commands header.

        :param address: Input address
        :param length: Input length
        :param cmd_tag: Command tag
        """
        self._address = address
        self._length = length
        self.cmd_tag = cmd_tag

    def info(self) -> str:
        """Get info of command.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def export(self) -> bytes:
        """Export command as bytes."""
        return pack(self.FORMAT, self.TAG, self.address, self.length, self.cmd_tag)

    @classmethod
    def header_parse(cls, cmd_tag: int, data: bytes, offset: int = 0) -> Tuple[int, int]:
        """Parse header command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :param cmd_tag: Information about command tag
        :raises SPSDKError: Raised if tag is not equal to required TAG
        :raises SPSDKError: Raised if cmd is not equal EnumCmdTag
        :return: Tuple
        """
        tag, address, length, cmd = unpack_from(cls.FORMAT, data, offset)
        if tag != cls.TAG:
            raise SPSDKError("TAG is not valid.")
        if cmd != cmd_tag:
            raise SPSDKError("Values are not same.")
        return address, length


########################################################################################################################
# Commands Classes version 3.1
########################################################################################################################
class CmdLoadBase(BaseCmd):
    """Base class for commands loading data."""

    HAS_MEMORY_ID_BLOCK = True

    def __init__(self, cmd_tag: int, address: int, data: bytes, memory_id: int = 0) -> None:
        """Constructor for command.

        :param cmd_tag: Command tag for the derived class
        :param address: Address for the load command
        :param data: Data to load
        :param memory_id: Memory ID
        """
        super().__init__(address=address, length=len(data), cmd_tag=cmd_tag)
        self.memory_id = memory_id
        self.data = data

    def export(self) -> bytes:
        """Export command as bytes."""
        data = super().export()
        if self.HAS_MEMORY_ID_BLOCK:
            data += pack("<4L", self.memory_id, 0, 0, 0)
        data += self.data
        data = align_block(data, alignment=16)
        return data

    def info(self) -> str:
        """Get info about the load command."""
        msg = f"{EnumCmdTag.name(self.cmd_tag)}: "
        if self.HAS_MEMORY_ID_BLOCK:
            msg += f"Address=0x{self.address:08X}, Length={self.length}, Memory ID={self.memory_id}"
        else:
            msg += f"Address=0x{self.address:08X}, Length={self.length}"
        return msg

    @classmethod
    def _extract_data(cls, data: bytes, offset: int = 0) -> Tuple[int, int, bytes, int, int]:
        tag, address, length, cmd = unpack_from(cls.FORMAT, data)
        memory_id = 0
        if tag != cls.TAG:
            raise SPSDKError(f"Invalid TAG, expected: {cls.TAG}")
        offset += BaseCmd.SIZE
        if cls.HAS_MEMORY_ID_BLOCK:
            memory_id, pad0, pad1, pad2 = unpack_from("<4L", data, offset=offset)
            if not pad0 == pad1 == pad2 == 0:
                raise SPSDKError("Invalid padding")
            offset += 16
        load_data = data[offset : offset + length]
        return address, length, load_data, cmd, memory_id

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdLoadBase":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdLoad
        :raises SPSDKError: Invalid cmd_tag was found
        """
        address, _, data, cmd_tag, memory_id = cls._extract_data(data, offset)
        if cmd_tag not in [
            EnumCmdTag.LOAD,
            EnumCmdTag.LOAD_CMAC,
            EnumCmdTag.LOAD_HASH_LOCKING,
            EnumCmdTag.LOAD_KEY_BLOB,
            EnumCmdTag.PROGRAM_FUSES,
            EnumCmdTag.PROGRAM_IFR,
        ]:
            raise SPSDKError(f"Invalid cmd_tag found: {cmd_tag}")
        if cls == CmdLoadBase:
            return cls(cmd_tag=cmd_tag, address=address, data=data, memory_id=memory_id)
        # pylint: disable=no-value-for-parameter
        return cls(address=address, data=data, memory_id=memory_id)  # type: ignore

    # pylint: disable=redundant-returns-doc
    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> MainCmd:
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        assert False


class CmdErase(BaseCmd):
    """Erase given address range. The erase will be rounded up to the sector size."""

    def __init__(self, address: int, length: int, memory_id: int = 0) -> None:
        """Constructor for command.

        :param address: Input address
        :param length: Input length
        :param memory_id: Memory ID
        """
        super().__init__(cmd_tag=EnumCmdTag.ERASE, address=address, length=length)
        self.memory_id = memory_id

    def info(self) -> str:
        """Get info of command."""
        return (
            f"ERASE: Address=0x{self.address:08X}, Length={self.length}, Memory ID={self.memory_id}"
        )

    def export(self) -> bytes:
        """Export command as bytes."""
        data = super().export()
        data += pack("<4L", self.memory_id, 0, 0, 0)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdErase":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdErase
        :raises SPSDKError: Invalid padding
        """
        address, length = cls.header_parse(data=data, offset=offset, cmd_tag=EnumCmdTag.ERASE)
        memory_id, pad0, pad1, pad2 = unpack_from("<4L", data, offset=offset + 16)
        if not pad0 == pad1 == pad2 == 0:
            raise SPSDKError("Invalid padding")
        return cls(address=address, length=length, memory_id=memory_id)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdErase":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["address"], 0)
        length = value_to_int(config["size"], 0)
        memory_id = value_to_int(config.get("memoryId", "0"), 0)
        return CmdErase(address=address, length=length, memory_id=memory_id)


class CmdLoad(CmdLoadBase):
    """Data to write follows the range header."""

    def __init__(self, address: int, data: bytes, memory_id: int = 0) -> None:
        """Constructor for command.

        :param address: Address for the load command
        :param data: Data to load
        :param memory_id: Memory ID
        """
        super().__init__(cmd_tag=EnumCmdTag.LOAD, address=address, data=data, memory_id=memory_id)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> Union["CmdLoad", "CmdLoadHashLocking", "CmdLoadCmac"]:
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        :raises SPSDKError: Invalid configuration field.
        """
        authentication = config.get("authentication")
        address = value_to_int(config["address"], 0)
        memory_id = value_to_int(config.get("memoryId", "0"), 0)
        if authentication == "hashlocking":
            data = load_binary(config["file"], search_paths=search_paths)
            return CmdLoadHashLocking.load_from_config(
                config, search_paths=search_paths
            )  # Backward compatibility
        if authentication == "cmac":
            data = load_binary(config["file"], search_paths=search_paths)
            return CmdLoadCmac.load_from_config(
                config, search_paths=search_paths
            )  # Backward compatibility
        # general non-authenticated load command
        if config.get("file"):
            data = load_binary(config["file"], search_paths=search_paths)
            return CmdLoad(address=address, data=data, memory_id=memory_id)
        if config.get("values"):
            values = [value_to_int(s, 0) for s in config["values"].split(",")]
            data = pack(f"<{len(values)}L", *values)
            return CmdLoad(address=address, data=data, memory_id=memory_id)

        raise SPSDKError(f"Unsupported LOAD command args: {config}")


class CmdExecute(BaseCmd):
    """Address will be the jump-to address."""

    def __init__(self, address: int) -> None:
        """Constructor for Command.

        :param address: Input address
        """
        super().__init__(cmd_tag=EnumCmdTag.EXECUTE, address=address, length=0)

    def info(self) -> str:
        """Get info of command."""
        return f"EXECUTE: Address=0x{self.address:08X}"

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdExecute":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdExecute
        """
        address, _ = cls.header_parse(data=data, offset=offset, cmd_tag=EnumCmdTag.EXECUTE)
        return cls(address=address)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdExecute":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["address"], 0)
        return CmdExecute(address=address)


class CmdCall(BaseCmd):
    """Address will be the address to jump."""

    def __init__(self, address: int) -> None:
        """Constructor for Command.

        :param address: Input address
        """
        super().__init__(cmd_tag=EnumCmdTag.CALL, address=address, length=0)

    def info(self) -> str:
        """Get info of command."""
        return f"CALL: Address=0x{self.address:08X}"

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdCall":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdCall
        """
        address, _ = cls.header_parse(data=data, offset=offset, cmd_tag=EnumCmdTag.CALL)
        return cls(address=address)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdCall":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["address"], 0)
        return CmdCall(address=address)


class CmdProgFuses(CmdLoadBase):
    """Address will be address of fuse register."""

    HAS_MEMORY_ID_BLOCK = False

    def __init__(self, address: int, data: bytes) -> None:
        """Constructor for Command.

        :param address: Input address
        :param data: Input data
        """
        super().__init__(cmd_tag=EnumCmdTag.PROGRAM_FUSES, address=address, data=data)
        self.length //= 4

    @classmethod
    def _extract_data(cls, data: bytes, offset: int = 0) -> Tuple[int, int, bytes, int, int]:
        tag, address, length, cmd = unpack_from(cls.FORMAT, data)
        length *= 4
        memory_id = 0
        if tag != cls.TAG:
            raise SPSDKError(f"Invalid TAG, expected: {cls.TAG}")
        offset += BaseCmd.SIZE
        if cls.HAS_MEMORY_ID_BLOCK:
            memory_id, pad0, pad1, pad2 = unpack_from("<4L", data, offset=offset)
            if pad0 != pad1 != pad2 != 0:
                raise SPSDKError("Invalid padding")
            offset += 16
        load_data = data[offset : offset + length]
        return address, length, load_data, cmd, memory_id

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdProgFuses":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdProgFuses
        """
        address, _, data, _, _ = cls._extract_data(data=data, offset=offset)
        return cls(address=address, data=data)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdProgFuses":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["address"], 0)
        fuses = [value_to_int(fuse, 0) for fuse in config["values"].split(",")]
        data = pack(f"<{len(fuses)}L", *fuses)
        return CmdProgFuses(address=address, data=data)


class CmdProgIfr(CmdLoadBase):
    """Address will be the address into the IFR region."""

    HAS_MEMORY_ID_BLOCK = False

    def __init__(self, address: int, data: bytes) -> None:
        """Constructor for Command.

        :param address: Input address
        :param data: Input data as bytes array
        """
        super().__init__(cmd_tag=EnumCmdTag.PROGRAM_IFR, address=address, data=data)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdProgIfr":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdProgFuses
        """
        address, _, data, _, _ = cls._extract_data(data=data, offset=offset)
        return cls(address=address, data=data)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdProgIfr":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["address"], 0)
        data = load_binary(config["file"], search_paths=search_paths)
        return CmdProgIfr(address=address, data=data)


class CmdLoadCmac(CmdLoadBase):
    """Load cmac. ROM is calculating cmac from loaded data."""

    def __init__(self, address: int, data: bytes, memory_id: int = 0) -> None:
        """Constructor for command.

        :param address: Address for the load command
        :param data: Data to load
        :param memory_id: Memory ID
        """
        super().__init__(
            cmd_tag=EnumCmdTag.LOAD_CMAC,
            address=address,
            data=data,
            memory_id=memory_id,
        )

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdLoadCmac":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        :raises SPSDKError: Invalid configuration field.
        """
        address = value_to_int(config["address"], 0)
        memory_id = value_to_int(config.get("memoryId", "0"), 0)

        data = load_binary(config["file"], search_paths=search_paths)
        return CmdLoadCmac(address=address, data=data, memory_id=memory_id)


class CmdCopy(BaseCmd):
    """Copy data from one place to another."""

    def __init__(
        self,
        address: int,
        length: int,
        destination_address: int = 0,
        memory_id_from: int = 0,
        memory_id_to: int = 0,
    ) -> None:
        """Constructor for command.

        :param address: Input address
        :param length: Input length
        :param destination_address: Destination address
        :param memory_id_from: Memory ID
        :param memory_id_to: Memory ID
        """
        super().__init__(cmd_tag=EnumCmdTag.COPY, address=address, length=length)
        self.destination_address = destination_address
        self.memory_id_from = memory_id_from
        self.memory_id_to = memory_id_to

    def info(self) -> str:
        """Get info of command."""
        return (
            f"COPY: Address=0x{self.address:08X}, Length={self.length}, "
            f"Destination address={self.destination_address}"
            f"Memory ID from={self.memory_id_from}, Memory ID to={self.memory_id_to}"
        )

    def export(self) -> bytes:
        """Export command as bytes."""
        data = super().export()
        data += pack("<4L", self.destination_address, self.memory_id_from, self.memory_id_to, 0)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdCopy":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdCopy
        :raises SPSDKError: Invalid padding
        """
        address, length = cls.header_parse(data=data, offset=offset, cmd_tag=EnumCmdTag.COPY)
        destination_address, memory_id_from, memory_id_to, pad0 = unpack_from(
            "<4L", data, offset=offset + 16
        )
        if pad0 != 0:
            raise SPSDKError("Invalid padding")
        return cls(
            address=address,
            length=length,
            destination_address=destination_address,
            memory_id_from=memory_id_from,
            memory_id_to=memory_id_to,
        )

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdCopy":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["addressFrom"], 0)
        length = value_to_int(config["size"], 0)
        destination_address = value_to_int(config["addressTo"], 0)
        memory_id_from = value_to_int(config["memoryIdFrom"], 0)
        memory_id_to = value_to_int(config["memoryIdTo"], 0)
        return CmdCopy(
            address=address,
            length=length,
            destination_address=destination_address,
            memory_id_from=memory_id_from,
            memory_id_to=memory_id_to,
        )


class CmdLoadHashLocking(CmdLoadBase):
    """Load hash. ROM is calculating hash."""

    def __init__(self, address: int, data: bytes, memory_id: int = 0) -> None:
        """Constructor for command.

        :param address: Address for the load command
        :param data: Data to load
        :param memory_id: Memory ID
        """
        super().__init__(
            cmd_tag=EnumCmdTag.LOAD_HASH_LOCKING,
            address=address,
            data=data,
            memory_id=memory_id,
        )

    def export(self) -> bytes:
        """Export command as bytes."""
        data = super().export()
        data += bytes(64)
        return data

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdLoadHashLocking":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        :raises SPSDKError: Invalid configuration field.
        """
        address = value_to_int(config["address"], 0)
        memory_id = value_to_int(config.get("memoryId", "0"), 0)

        data = load_binary(config["file"], search_paths=search_paths)
        return CmdLoadHashLocking(address=address, data=data, memory_id=memory_id)


class CmdLoadKeyBlob(BaseCmd):
    """Load key blob."""

    FORMAT = "<L2H2L"

    class _KeyWraps(BuiltinEnum):
        """KeyWrap IDs used by the CmdLoadKeyBlob command."""

        NXP_CUST_KEK_INT_SK = 16
        NXP_CUST_KEK_EXT_SK = 17

    class _KeyWrapsV2(BuiltinEnum):
        """KeyWrap IDs used by the CmdLoadKeyBlob command for mcxnx family."""

        NXP_CUST_KEK_INT_SK = 18
        NXP_CUST_KEK_EXT_SK = 19

    class KeyTypes(BuiltinEnum):
        """KeyTypes for NXP_CUST_KEK_INT_SK, NXP_CUST_KEK_EXT_SK."""

        NXP_CUST_KEK_INT_SK = 1
        NXP_CUST_KEK_EXT_SK = 2

    @classmethod
    def get_key_id(cls, family: str, key_name: KeyTypes) -> int:
        """Get key ID based on family and key name.

        :param family: chip family
        :param key_name: NXP_CUST_KEK_INT_SK or NXP_CUST_KEK_EXT_SK
        :return: integer value representing key
        """
        key_wraps = cls._KeyWrapsV2 if "mcxn" in family.lower() else cls._KeyWraps
        return key_wraps[key_name.name].value

    def __init__(self, offset: int, data: bytes, key_wrap_id: int) -> None:
        """Constructor for command.

        :param offset: Input offset
        :param key_wrap_id: Key wrap ID (NXP_CUST_KEK_INT_SK = 16, NXP_CUST_KEK_EXT_SK = 17)
        :param data: Wrapped key blob
        """
        super().__init__(cmd_tag=EnumCmdTag.LOAD_KEY_BLOB, address=offset, length=len(data))
        self.key_wrap_id = key_wrap_id
        self.data = data

    def info(self) -> str:
        """Get info of command."""
        return f"LOAD_KEY_BLOB: Offset=0x{self.address:08X}, Length={self.length}, Key wrap ID={self.key_wrap_id}"

    def export(self) -> bytes:
        """Export command as bytes."""
        result_data = pack(
            self.FORMAT,
            self.TAG,
            self.address,
            self.key_wrap_id,
            self.length,
            self.cmd_tag,
        )
        result_data += self.data

        result_data = align_block(data=result_data, alignment=16, padding=0)
        return result_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdLoadKeyBlob":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdLoadKeyBlob
        """
        tag, cmpa_offset, key_wrap_id, length, cmd = unpack_from(  # pylint: disable=unused-variable
            cls.FORMAT, data, offset
        )
        key_blob_data = unpack_from(f"<{length}s", data, offset + cls.SIZE)[0]
        return cls(offset=cmpa_offset, key_wrap_id=key_wrap_id, data=key_blob_data)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdLoadKeyBlob":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        data = load_binary(config["file"], search_paths=search_paths)
        offset = value_to_int(config["offset"], 0)
        key_wrap_name = config["wrappingKeyId"]
        family = config["family"]

        key_wrap_id = cls.get_key_id(family, cls.KeyTypes[key_wrap_name])
        return CmdLoadKeyBlob(offset=offset, data=data, key_wrap_id=key_wrap_id)


class CmdConfigureMemory(BaseCmd):
    """Configure memory."""

    def __init__(self, address: int, memory_id: int = 0) -> None:
        """Constructor for command.

        :param address: Input address
        :param memory_id: Memory ID
        """
        super().__init__(address=address, length=0, cmd_tag=EnumCmdTag.CONFIGURE_MEMORY)
        self.memory_id = memory_id

    def info(self) -> str:
        """Get info of command."""
        return f"CONFIGURE_MEMORY: Address=0x{self.address:08X}, Memory ID={self.memory_id}"

    def export(self) -> bytes:
        """Export command as bytes."""
        return pack(self.FORMAT, self.TAG, self.memory_id, self.address, self.cmd_tag)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdConfigureMemory":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdConfigureMemory
        """
        memory_id, address = cls.header_parse(
            cmd_tag=EnumCmdTag.CONFIGURE_MEMORY, data=data, offset=offset
        )
        return cls(address=address, memory_id=memory_id)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdConfigureMemory":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        memory_id = value_to_int(config["memoryId"], 0)
        return CmdConfigureMemory(
            address=value_to_int(config["configAddress"], 0), memory_id=memory_id
        )


class CmdFillMemory(BaseCmd):
    """Fill memory range by pattern."""

    def __init__(self, address: int, length: int, pattern: int) -> None:
        """Constructor for command.

        :param address: Input address
        :param length: Input length
        :param pattern: Pattern for fill memory with
        """
        super().__init__(cmd_tag=EnumCmdTag.FILL_MEMORY, address=address, length=length)
        self.pattern = pattern

    def info(self) -> str:
        """Get info of command."""
        return f"FILL_MEMORY: Address=0x{self.address:08X}, Length={self.length}, PATTERN={hex(self.pattern)}"

    def export(self) -> bytes:
        """Export command as bytes."""
        data = super().export()
        data += pack("<4L", self.pattern, 0, 0, 0)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdFillMemory":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdErase
        :raises SPSDKError: Invalid padding
        """
        address, length = cls.header_parse(data=data, offset=offset, cmd_tag=EnumCmdTag.FILL_MEMORY)
        pattern, pad0, pad1, pad2 = unpack_from("<4L", data, offset=offset + 16)
        if pad0 != pad1 != pad2 != 0:
            raise SPSDKError("Invalid padding")
        return cls(address=address, length=length, pattern=pattern)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdFillMemory":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        address = value_to_int(config["address"], 0)
        length = value_to_int(config["size"], 0)
        pattern = value_to_int(config["pattern"], 0)
        return CmdFillMemory(address=address, length=length, pattern=pattern)


class CmdFwVersionCheck(BaseCmd):
    """Check counter value with stored value, if values are not same, SB file is rejected."""

    class CounterID(Enum):
        """Counter IDs used by the CmdFwVersionCheck command."""

        NONE = (0, "none")
        NONSECURE = (1, "nonsecure")
        SECURE = (2, "secure")
        RADIO = (3, "radio")
        SNT = (4, "snt")
        BOOTLOADER = (5, "bootloader")

    def __init__(self, value: int, counter_id: int) -> None:
        """Constructor for command.

        :param value: Input value
        :param counter_id: Counter ID (NONSECURE = 1, SECURE = 2)
        """
        super().__init__(address=0, length=0, cmd_tag=EnumCmdTag.FW_VERSION_CHECK)
        self.value = value
        self.counter_id = counter_id

    def info(self) -> str:
        """Get info of command."""
        return f"FW_VERSION_CHECK: Value={self.value}, Counter ID={self.counter_id}"

    def export(self) -> bytes:
        """Export command as bytes."""
        return pack(self.FORMAT, self.TAG, self.value, self.counter_id, self.cmd_tag)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdFwVersionCheck":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :return: CmdFwVersionCheck
        """
        value, counter_id = cls.header_parse(
            data=data, offset=offset, cmd_tag=EnumCmdTag.FW_VERSION_CHECK
        )
        return cls(value=value, counter_id=counter_id)

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdFwVersionCheck":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        """
        value = value_to_int(config["value"], 0)
        counter_id_str = config["counterId"]
        counter_id = CmdFwVersionCheck.CounterID[counter_id_str]
        return CmdFwVersionCheck(value=value, counter_id=counter_id)


class CmdSectionHeader(MainCmd):
    """Create section header."""

    FORMAT = "<4L"
    SIZE = calcsize(FORMAT)

    def __init__(self, length: int, section_uid: int = 1, section_type: int = 1) -> None:
        """Constructor for Commands section.

        :param section_uid: Input uid
        :param section_type: Input type
        :param length: Input length
        """
        self.section_uid = section_uid
        self.section_type = section_type
        self.length = length
        self._pad = 0

    def info(self) -> str:
        """Get info of Section header."""
        return f"Section header: UID=0x{self.section_uid:08X}, Type={self.section_type}, Length={self.length}"

    def export(self) -> bytes:
        """Export command as bytes."""
        return pack(self.FORMAT, self.section_uid, self.section_type, self.length, self._pad)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdSectionHeader":
        """Parse command from bytes array.

        :param data: Input data as bytes array
        :param offset: The offset of input data
        :raises SPSDKError: Raised when FORMAT is bigger than length of the data without offset
        :return: CmdSectionHeader
        """
        if calcsize(cls.FORMAT) > len(data) - offset:
            raise SPSDKError("FORMAT is bigger than length of the data without offset!")
        section_uid, section_type, length, _ = unpack_from(cls.FORMAT, data, offset)
        return cls(section_uid=section_uid, section_type=section_type, length=length)

    # pylint: disable=redundant-returns-doc
    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CmdSectionHeader":
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Command object loaded from configuration.
        :raises SPSDKError: This situation cannot raise (the function here is just MYPY/PYLINT checks).
        """
        raise SPSDKError("Section header cannot be loaded from configuration.")


TAG_TO_CLASS: Mapping[int, Type[BaseCmd]] = {
    EnumCmdTag.ERASE: CmdErase,
    EnumCmdTag.LOAD: CmdLoad,
    EnumCmdTag.EXECUTE: CmdExecute,
    EnumCmdTag.CALL: CmdCall,
    EnumCmdTag.PROGRAM_FUSES: CmdProgFuses,
    EnumCmdTag.PROGRAM_IFR: CmdProgIfr,
    EnumCmdTag.LOAD_CMAC: CmdLoadCmac,
    EnumCmdTag.COPY: CmdCopy,
    EnumCmdTag.LOAD_HASH_LOCKING: CmdLoadHashLocking,
    EnumCmdTag.LOAD_KEY_BLOB: CmdLoadKeyBlob,
    EnumCmdTag.CONFIGURE_MEMORY: CmdConfigureMemory,
    EnumCmdTag.FILL_MEMORY: CmdFillMemory,
    EnumCmdTag.FW_VERSION_CHECK: CmdFwVersionCheck,
}

CFG_NAME_TO_CLASS: Mapping[str, Type[BaseCmd]] = {
    "erase": CmdErase,
    "load": CmdLoad,
    "execute": CmdExecute,
    "call": CmdCall,
    "programFuses": CmdProgFuses,
    "programIFR": CmdProgIfr,
    "loadCMAC": CmdLoadCmac,
    "copy": CmdCopy,
    "loadHashLocking": CmdLoadHashLocking,
    "loadKeyBlob": CmdLoadKeyBlob,
    "configureMemory": CmdConfigureMemory,
    "fillMemory": CmdFillMemory,
    "checkFwVersion": CmdFwVersionCheck,
}


########################################################################################################################
# Command parser from raw data
########################################################################################################################
def parse_command(data: bytes, offset: int = 0) -> object:
    """Parse command from bytes array.

    :param data: Input data as bytes array
    :param offset: The offset of input data
    :raises SPSDKError: Raised when tag is not in cmd_class
    :raises SPSDKError: Raised when tag is invalid
    :return: object
    """
    #  verify that first 4 bytes of frame are 55aaaa55
    tag = unpack_from("<L", data, offset=offset)[0]
    if tag != BaseCmd.TAG:
        raise SPSDKError("Invalid tag.")
    cmd_tag = unpack_from("<L", data, offset=offset + 12)[0]
    if cmd_tag not in TAG_TO_CLASS:
        raise SPSDKError(f"Invalid command tag: {cmd_tag}")
    return TAG_TO_CLASS[cmd_tag].parse(data, offset)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains XMCD (External Memory Configuration Data) related code."""
import os

from spsdk import SPSDK_DATA_FOLDER

XMCD_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "image", "xmcd")
XMCD_SCH_FILE: str = os.path.join(XMCD_DATA_FOLDER, "sch_xmcd.yaml")
XMCD_DATABASE_FILE: str = os.path.join(XMCD_DATA_FOLDER, "database.yaml")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides support for Protected Flash Region areas (CMPA, CFPA)."""
import copy
import logging
import math
import os
from typing import Any, Dict, List, Optional, Union

from ruamel.yaml.comments import CommentedMap as CM

from spsdk import SPSDK_DATA_FOLDER, SPSDKError
from spsdk import __version__ as spsdk_version
from spsdk.crypto import PublicKey, ec, rsa
from spsdk.utils.crypto.abstract import BackendClass
from spsdk.utils.crypto.backend_openssl import openssl_backend
from spsdk.utils.crypto.rkht import RKHT
from spsdk.utils.exceptions import SPSDKRegsErrorRegisterNotFound
from spsdk.utils.misc import load_configuration, value_to_int
from spsdk.utils.reg_config import RegConfig
from spsdk.utils.registers import Registers, RegsRegister

from . import PFR_DATA_FOLDER
from .exceptions import (
    SPSDKPfrConfigError,
    SPSDKPfrConfigReadError,
    SPSDKPfrError,
    SPSDKPfrRotkhIsNotPresent,
)

logger = logging.getLogger(__name__)


class PfrConfiguration:
    """Class to open PFR configuration file a get basic configuration."""

    def __init__(
        self,
        config: Optional[Union[str, dict, "PfrConfiguration"]] = None,
        device: Optional[str] = None,
        revision: Optional[str] = None,
        cfg_type: Optional[str] = None,
    ) -> None:
        """Open config PFR file.

        :param config: Filename or dictionary with PFR settings, defaults to None
        :param device: If needed it could be used to override device from settings, defaults to ""
        :param revision: If needed it could be used to override revision from settings, defaults to ""
        :param cfg_type: If needed it could be used to override PFR type from settings, defaults to ""
        """
        self.device = device
        self.revision = revision
        self.type = cfg_type
        self.settings: Optional[Union[CM, dict]] = None

        if isinstance(config, (str, os.PathLike, dict)):
            self.set_config(config)

        if isinstance(config, PfrConfiguration):
            if not self.device:
                self.device = config.device
            if not self.revision:
                self.revision = config.revision
            if not self.type:
                self.type = config.type
            if config.settings:
                self.settings = config.settings.copy()

    @staticmethod
    def _detect_obsolete_style_of_settings(data: Union[CM, dict]) -> bool:
        """Detect obsolete style of configuration.

        :param data: As old JSON style as new YML style of settings data.
        :return: True if obsolete style is detected.
        """
        if len(data) == 0:
            return False

        for key in data.keys():
            if isinstance(data[key], (str, int)):
                return True
            if isinstance(data[key], dict):
                first_key = list(data[key].keys())[0]
                if first_key not in ("value", "bitfields", "name"):
                    return True

        return False

    def _get_yml_style_of_settings(self, data: Union[CM, dict]) -> Union[CM, dict]:
        """Get unified YML style of settings.

        :param data: As old JSON style as new YML style of settings data.
        :return: New YML style of data.
        """
        if not self._detect_obsolete_style_of_settings(data):
            return data

        yml_style: Dict[str, Union[str, int, dict]] = {}
        for key, val in data.items():
            if isinstance(val, (str, int)):
                yml_style[key] = {"value": val}
            if isinstance(val, dict):
                bitfields = {}
                for key_b, val_b in val.items():
                    bitfields[key_b] = val_b
                yml_style[key] = {"bitfields": bitfields}

        return yml_style

    def set_config_dict(
        self,
        data: Union[CM, dict],
        device: Optional[str] = None,
        revision: Optional[str] = None,
        cfg_type: Optional[str] = None,
    ) -> None:
        """Apply configuration dictionary.

        The function accepts as dictionary as from commented map.

        :param data: Settings of PFR.
        :param device: If needed it could be used to override device from settings, defaults to ""
        :param revision: If needed it could be used to override revision from settings, defaults to ""
        :param cfg_type: If needed it could be used to override PFR type from settings, defaults to ""
        :raises SPSDKPfrConfigReadError: Invalid YML file.
        """
        if data is None or len(data) == 0:
            raise SPSDKPfrConfigReadError("Empty YAML configuration.")

        try:
            description = data.get("description", data)
            self.device = device or description.get("device", None)
            self.revision = revision or description.get("revision", None)
            self.type = cfg_type or description.get("type", None)
            self.settings = self._get_yml_style_of_settings(data["settings"])

        except KeyError as exc:
            raise SPSDKPfrConfigReadError("Missing fields in YAML configuration.") from exc

    def set_config(self, config: Union[str, CM, dict]) -> None:
        """Apply configuration from file.

        :param config: Name of configuration file or Commented map.
        :raises SPSDKPfrConfigReadError: The configuration file cannot be loaded.
        """
        if isinstance(config, (CM, dict)):
            self.set_config_dict(config)
        else:
            try:
                data = load_configuration(config)
            except SPSDKError as exc:
                raise SPSDKPfrConfigReadError(str(exc)) from exc
            self.set_config_dict(data)

    def get_yaml_config(self, data: CM, indent: int = 0) -> CM:
        """Return YAML configuration In PfrConfiguration format.

        :param data: The registers settings data.
        :param indent: YAML start indent.
        :return: YAML PFR configuration in commented map(ordered dict).
        :raises SPSDKError: When there is no device found
        :raises SPSDKError: When there is no type found
        """
        if not self.device:
            raise SPSDKError("Device not found")
        if not self.type:
            raise SPSDKError("Type not found")
        res_data = CM()

        res_data.yaml_set_start_comment(
            f"NXP {self.device} PFR {self.type} configuration", indent=indent
        )

        description = CM()
        description.insert(1, "device", self.device, comment="The NXP device name.")
        description.insert(2, "revision", self.revision, comment="The NXP device revision.")
        description.insert(
            3, "type", self.type.upper(), comment="The PFR type (CMPA, CFPA) or IFR type."
        )
        description.insert(4, "version", spsdk_version, comment="The SPSDK tool version.")

        res_data.insert(
            1,
            "description",
            description,
            comment=f"The {self.type} configuration description.",
        )
        res_data.insert(2, "settings", data, comment=f"The {self.type} registers configuration.")
        return res_data

    def is_invalid(self) -> Optional[str]:
        """Validate configuration.

        :return: None if configuration is valid, otherwise description string what is invalid.
        """
        if not self.device:
            return "The device is NOT specified!"
        if not self.type:
            return "The PFR type (CMPA/CFPA) is NOT specified!"

        return None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PfrConfiguration):
            return False

        return vars(self) == vars(other)


class BaseConfigArea:
    """Base for CMPA and CFPA classes."""

    CONFIG_DIR = PFR_DATA_FOLDER
    CONFIG_FILE = "database.yaml"
    BINARY_SIZE = 512
    ROTKH_SIZE = 32
    ROTKH_REGISTER = "ROTKH"
    MARK = b"SEAL"
    DESCRIPTION = "Base Config Area"
    IMAGE_PREFILL_PATTERN = 0

    def __init__(
        self,
        device: Optional[str] = None,
        revision: Optional[str] = None,
        user_config: Optional[PfrConfiguration] = None,
        raw: bool = False,
    ) -> None:
        """Initialize an instance.

        :param device: device to use, list of supported devices is available via 'devices' method
        :param revision: silicon revision, if not specified, the latest is being used
        :param user_config: PfrConfiguration with user configuration to use with initialization
        :param raw: When set the computed fields from configuration will be applied
        :raises SPSDKError: When no device is provided
        :raises SPSDKError: When no device is not supported
        :raises SPSDKError: When there is invalid revision
        """
        if not (device or user_config):
            raise SPSDKError("No device provided")
        self.config = self._load_config()
        # either 'device' or 'user_config' IS defined! Mypy doesn't understand the check above
        self.device = device or user_config.device  # type: ignore

        if self.device not in self.config.devices.device_names:
            raise SPSDKError(f"Device '{self.device}' is not supported")
        self.revision = revision or (user_config.revision if user_config else "latest")
        revisions = self.config.devices.get_by_name(self.device).revisions  # type: ignore
        if not self.revision or self.revision == "latest":
            self.revision = revisions.get_latest().name
            logger.info(
                f"The silicon revision is not specified, the latest: '{self.revision}' has been used."
            )

        if self.revision not in revisions.revision_names:
            raise SPSDKError(f"Invalid revision '{self.revision}' for '{self.device}'")
        self.registers = Registers(self.device, base_endianness="little")  # type: ignore
        self.registers.load_registers_from_xml(
            xml=self.config.get_data_file(self.device, self.revision),  # type: ignore
            filter_reg=self.config.get_ignored_registers(self.device),
            grouped_regs=self.config.get_grouped_registers(self.device),
        )

        # Set the computed field handler
        for reg, fields in self.config.get_computed_fields(self.device).items():
            reg_obj = self.registers.find_reg(reg)
            reg_obj.add_setvalue_hook(self.reg_computed_fields_handler, fields)

        self.user_config = PfrConfiguration(
            config=user_config,
            device=self.device,
            revision=self.revision,
            cfg_type=self.__class__.__name__,
        )

        if self.user_config.settings:
            self.set_config(self.user_config, raw=raw)

    def reg_computed_fields_handler(self, val: bytes, context: Any) -> bytes:
        """Recalculate all fields for given register value.

        :param val: Input register value.
        :param context: The method context (fields).
        :return: recomputed value.
        :raises SPSDKPfrError: Raises when the computing routine is not found.
        """
        fields: dict = context
        for method in fields.values():
            if hasattr(self, method):
                method_ref = getattr(self, method)
                val = method_ref(val)
            else:
                raise SPSDKPfrError(f"The '{method}' compute function doesn't exists.")

        return val

    @staticmethod
    def pfr_reg_inverse_high_half(val: int) -> int:
        """Function that inverse low 16-bits of register value to high 16 bits.

        :param val: Input current reg value.
        :return: Returns the complete register value with updated higher half field.
        """
        ret = val & 0xFFFF
        ret |= (ret ^ 0xFFFF) << 16
        return ret

    @staticmethod
    def pfr_reg_inverse_lower_8_bits(val: int) -> int:
        """Function that inverse lower 8-bits of register value to 8-16 bits.

        :param val: Input current reg value.
        :return: Returns the complete register value with updated 8-16 bit field.
        """
        ret = val & 0xFFFF_00FF
        inverse = (val & 0xFF) ^ 0xFF
        ret |= inverse << 8
        return ret

    @classmethod
    def _load_config(cls) -> RegConfig:
        """Loads the PFR block configuration file.

        :return: PFR block configuration database.
        """
        return RegConfig(os.path.join(cls.CONFIG_DIR, cls.CONFIG_FILE))

    @classmethod
    def devices(cls) -> List[str]:
        """Classmethod to get list of supported devices.

        :return: List of supported devices.
        """
        config = cls._load_config()
        return config.devices.device_names

    def _get_registers(self) -> List[RegsRegister]:
        """Get a list of all registers.

        :return: List of PFR configuration registers.
        """
        exclude = self.config.get_ignored_registers(self.device)
        return self.registers.get_registers(exclude)

    def set_config(self, config: PfrConfiguration, raw: bool = False) -> None:
        """Apply configuration from file.

        :param config: PFR configuration.
        :param raw: When set all (included computed fields) configuration will be applied.
        :raises SPSDKError: When device is not provided.
        :raises SPSDKError: When revision is not provided.
        :raises SPSDKPfrConfigError: Invalid config file.
        """
        if not self.device:
            raise SPSDKError("No device provided")
        if not self.revision:
            raise SPSDKError("No revision provided")

        if config.device != self.device:
            raise SPSDKPfrConfigError(
                f"Invalid device in configuration. {self.device} != {config.device}"
            )
        if not config.revision or config.revision in ("latest", ""):
            config.revision = (
                self.config.devices.get_by_name(self.device).revisions.get_latest().name
            )
            logger.warning(
                f"The configuration file doesn't contains silicon revision,"
                f" the latest: '{config.revision}' has been used."
            )
        if config.revision != self.revision:
            raise SPSDKPfrConfigError(
                f"Invalid revision in configuration. {self.revision} != {config.revision}"
            )
        if config.type and config.type.upper() != self.__class__.__name__:
            raise SPSDKPfrConfigError(
                f"Invalid configuration type. {self.__class__.__name__} != {config.type}"
            )

        if not config.settings:
            raise SPSDKPfrConfigError("Missing configuration of PFR fields!")

        computed_regs = []
        computed_regs.extend(self.config.get_ignored_registers(self.device))
        if not raw:
            computed_regs.extend(list(self.config.get_computed_registers(self.device).keys()))
        computed_fields = None if raw else self.config.get_computed_fields(self.device)

        self.registers.load_yml_config(config.settings, computed_regs, computed_fields)
        if not raw:
            # Just update only configured registers
            exclude_hooks = []
            if not self.config.get_value("mandatory_computed_regs", self.device):
                exclude_hooks.extend(
                    list(set(self.registers.get_reg_names()) - set(config.settings.keys()))
                )
            self.registers.run_hooks(exclude_hooks)

    def get_yaml_config(
        self, exclude_computed: bool = True, diff: bool = False, indent: int = 0
    ) -> CM:
        """Return YAML configuration from loaded registers.

        :param exclude_computed: Omit computed registers and fields.
        :param diff: Get only configuration with difference value to reset state.
        :param indent: YAML start indent.
        :return: YAML PFR configuration in commented map(ordered dict).
        """
        computed_regs = (
            None
            if not exclude_computed
            else list(self.config.get_computed_registers(self.device).keys())
        )
        computed_fields = (
            None if not exclude_computed else self.config.get_computed_fields(self.device)
        )
        ignored_fields = self.config.get_ignored_fields(self.device)

        data = self.registers.create_yml_config(
            computed_regs, computed_fields, ignored_fields, diff, indent + 2
        )
        return self.user_config.get_yaml_config(data, indent)

    def generate_config(self, exclude_computed: bool = True) -> CM:
        """Generate configuration structure for user configuration.

        :param exclude_computed: Exclude computed fields, defaults to True.
        :return: YAML commented map with PFR configuration  in reset state.
        """
        # Create own copy to keep self as is and get reset values by standard YML output
        copy_of_self = copy.deepcopy(self)
        copy_of_self.registers.reset_values()

        return copy_of_self.get_yaml_config(exclude_computed)

    def _calc_rotkh(self, keys: List[PublicKey]) -> bytes:
        """Calculate ROTKH (Root Of Trust Key Hash).

        :param keys: List of Keys to compute ROTKH.
        :return: Value of ROTKH with right width.
        :raises SPSDKPfrError: Algorithm width doesn't fit into ROTKH field.
        """
        # the data structure use for computing final ROTKH is 4*32B long
        # 32B is a hash of individual keys
        # 4 is the max number of keys, if a key is not provided the slot is filled with '\x00'
        # The LPC55S3x has two options to compute ROTKH, so it's needed to be
        # detected the right algorithm and mandatory warn user about this selection because
        # it's MUST correspond to settings in eFuses!
        reg_rotkh = self.registers.find_reg("ROTKH")
        rkht = RKHT(keys=keys, keys_cnt=4, min_keys_cnt=1)
        rkht.validate()

        if rkht.hash_algorithm_size > reg_rotkh.width:
            raise SPSDKPfrError("The ROTKH field is smaller than used algorithm width.")

        return rkht.rotkh().ljust(reg_rotkh.width // 8, b"\x00")

    def _get_seal_start_address(self) -> int:
        """Function returns start of seal fields for the device.

        :return: Start of seals fields.
        :raises SPSDKError: When 'seal_start_address' in database.yaml can not be found
        """
        start = self.config.get_seal_start_address(self.device)
        if not start:
            raise SPSDKError("Can't find 'seal_start_address' in database.yaml")
        return self.registers.find_reg(start).offset

    def _get_seal_count(self) -> int:
        """Function returns seal count for the device.

        :return: Count of seals fields.
        :raises SPSDKError: When 'seal_count' in database.yaml can not be found
        """
        count = self.config.get_seal_count(self.device)
        if not count:
            raise SPSDKError("Can't find 'seal_count' in database.yaml")
        return value_to_int(count)

    def export(self, add_seal: bool = False, keys: Optional[List[PublicKey]] = None) -> bytes:
        """Generate binary output.

        :param add_seal: The export is finished in the PFR record by seal.
        :param keys: List of Keys to compute ROTKH field.
        :return: Binary block with PFR configuration(CMPA or CFPA).
        :raises SPSDKPfrRotkhIsNotPresent: This PFR block doesn't contains ROTKH field.
        :raises SPSDKError: The size of data is {len(data)}, is not equal to {self.BINARY_SIZE}.
        """
        if keys:
            try:
                # ROTKH may or may not be present, derived class defines its presence
                rotkh_reg = self.registers.find_reg(self.ROTKH_REGISTER)
                rotkh_data = self._calc_rotkh(keys)
                rotkh_reg.set_value(rotkh_data, True)
            except SPSDKRegsErrorRegisterNotFound as exc:
                raise SPSDKPfrRotkhIsNotPresent(
                    "This device doesn't contain ROTKH register!"
                ) from exc

        data = bytearray([self.IMAGE_PREFILL_PATTERN] * self.BINARY_SIZE)
        for reg in self._get_registers():
            data[reg.offset : reg.offset + reg.width // 8] = reg.get_bytes_value()

        if add_seal:
            seal_start = self._get_seal_start_address()
            seal_count = self._get_seal_count()
            data[seal_start : seal_start + seal_count * 4] = self.MARK * seal_count

        if len(data) != self.BINARY_SIZE:
            raise SPSDKError(f"The size of data is {len(data)}, is not equal to {self.BINARY_SIZE}")
        return bytes(data)

    def parse(self, data: bytes) -> None:
        """Parse input binary data to registers.

        :param data: Input binary data of PFR block.
        """
        for reg in self.registers.get_registers(include_group_regs=False):
            value = data[reg.offset : reg.offset + reg.width // 8]
            reg.set_bytes_value(value)


class CMPA(BaseConfigArea):
    """Customer Manufacturing Configuration Area."""

    CONFIG_DIR = os.path.join(BaseConfigArea.CONFIG_DIR, "cmpa")
    DESCRIPTION = "Customer Manufacturing Programmable Area"


class CFPA(BaseConfigArea):
    """Customer In-Field Configuration Area."""

    CONFIG_DIR = os.path.join(BaseConfigArea.CONFIG_DIR, "cfpa")
    DESCRIPTION = "Customer In-field Programmable Area"


class ROMCFG(BaseConfigArea):
    """Information flash region - ROMCFG."""

    CONFIG_DIR = os.path.join(os.path.join(SPSDK_DATA_FOLDER, "ifr"))
    BINARY_SIZE = 304
    IMAGE_PREFILL_PATTERN = 0xFF


def calc_pub_key_hash(
    public_key: PublicKey,
    backend: BackendClass = openssl_backend,
    sha_width: int = 256,
) -> bytes:
    """Calculate a hash out of public key's exponent and modulus in RSA case, X/Y in EC.

    :param public_key: List of public keys to compute hash from.
    :param backend: Crypto subsystem backend.
    :param sha_width: Used hash algorithm.
    :raises SPSDKError: Unsupported public key type
    :return: Computed hash.
    """
    if isinstance(public_key, rsa.RSAPublicKey):
        n_1 = public_key.public_numbers().e  # type: ignore # MyPy is unable to pickup the class member
        n1_len = math.ceil(n_1.bit_length() / 8)
        n_2 = public_key.public_numbers().n  # type: ignore # MyPy is unable to pickup the class member
        n2_len = math.ceil(n_2.bit_length() / 8)
    elif isinstance(public_key, ec.EllipticCurvePublicKey):
        n_1 = public_key.public_numbers().y  # type: ignore # MyPy is unable to pickup the class member
        n1_len = sha_width // 8
        n_2 = public_key.public_numbers().x  # type: ignore # MyPy is unable to pickup the class member
        n2_len = sha_width // 8
    else:
        raise SPSDKError(f"Unsupported key type: {type(public_key)}")

    n1_bytes = n_1.to_bytes(n1_len, "big")
    n2_bytes = n_2.to_bytes(n2_len, "big")

    return backend.hash(n2_bytes + n1_bytes, algorithm=f"sha{sha_width}")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module is used to generate public/private key and generating debug credential file."""
import logging
import os
import sys
from typing import List

import click

from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import spsdk_apps_common_options
from spsdk.apps.utils.utils import catch_spsdk_error, check_file_exists
from spsdk.crypto import (
    ec,
    generate_ecc_private_key,
    generate_ecc_public_key,
    generate_rsa_private_key,
    generate_rsa_public_key,
    save_ecc_private_key,
    save_ecc_public_key,
    save_rsa_private_key,
    save_rsa_public_key,
)

logger = logging.getLogger(__name__)


def get_list_of_supported_keys() -> List[str]:
    """Generate list with list of supported key types.

    :return: List of supported key types.
    """
    ret = ["rsa2048", "rsa3072", "rsa4096"]
    ret.extend(ec._CURVE_TYPES.keys())  # type: ignore # pylint: disable=protected-access

    return ret


@click.command(name="nxpkeygen", no_args_is_help=True)
@spsdk_apps_common_options
@click.option(
    "-k",
    "--key-type",
    type=click.Choice(get_list_of_supported_keys(), case_sensitive=False),
    metavar="KEY-TYPE",
    default="RSA2048",
    help=f"""\b
        Set of the supported key types. Default is RSA2048.

        Note: NXP DAT protocol is using encryption keys by this table:

        NXP Protocol Version                Encryption Type
            1.0                                 RSA 2048
            1.1                                 RSA 4096
            2.0                                 SECP256R1
            2.1                                 SECP384R1
            2.2                                 SECP521R1

        All possible options:
        {", ".join(get_list_of_supported_keys())}.
        """,
)
@click.option(
    "--password",
    "password",
    metavar="PASSWORD",
    help="Password with which the output file will be encrypted. "
    "If not provided, the output will be unencrypted.",
)
@click.argument("path", type=click.Path(file_okay=True, resolve_path=True))
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Force overwriting of an existing file.",
)
def main(log_level: int, key_type: str, path: str, password: str, force: bool) -> int:
    """NXP Key Generator Tool.

    !!! The NXPKEYGEN tool is deprecated, use new NXPCRYPTO tool from SPSDK for new projects !!!

    \b
    PATH    - output file path, where the key pairs (private and public key) will be stored.
              Each key will be stored in separate file (.pub and .pem).
    """
    spsdk_logger.install(level=log_level)

    key_param = key_type.lower().strip()
    is_rsa = "rsa" in key_param

    check_file_exists(path, force)
    pub_key_path = os.path.splitext(path)[0] + ".pub"
    check_file_exists(pub_key_path, force)

    if is_rsa:
        logger.info("Generating RSA private key...")
        priv_key_rsa = generate_rsa_private_key(key_size=int(key_param.replace("rsa", "")))
        logger.info("Generating RSA corresponding public key...")
        pub_key_rsa = generate_rsa_public_key(priv_key_rsa)
        logger.info("Saving RSA key pair...")
        save_rsa_private_key(priv_key_rsa, path, password if password else None)
        save_rsa_public_key(pub_key_rsa, pub_key_path)
    else:
        logger.info("Generating ECC private key...")
        priv_key_ec = generate_ecc_private_key(curve_name=key_param)
        logger.info("Generating ECC public key...")
        pub_key_ec = generate_ecc_public_key(priv_key_ec)
        logger.info("Saving ECC key pair...")
        save_ecc_private_key(priv_key_ec, path, password if password else None)
        save_ecc_public_key(pub_key_ec, pub_key_path)
    click.echo(f"The key pair has been created: {(pub_key_path)}, {(path)}")
    return 0


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing functionality of srktool, dcdgen, mkimage and other similar tools."""

import os

from spsdk import SPSDK_DATA_FOLDER

IMG_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "image")
TZ_SCH_FILE: str = os.path.join(IMG_DATA_FOLDER, "sch_tz.yaml")
MBIMG_SCH_FILE: str = os.path.join(IMG_DATA_FOLDER, "sch_mbimg.yaml")

from .bee import BeeFacRegion, BeeKIB, BeeProtectRegionBlock, BeeRegionHeader
from .commands import (
    CmdAuthData,
    CmdCheckData,
    CmdInitialize,
    CmdInstallKey,
    CmdNop,
    CmdSet,
    CmdUnlock,
    CmdUnlockCAAM,
    CmdUnlockOCOTP,
    CmdUnlockSNVS,
    CmdWriteData,
    EnumAuthDat,
    EnumCertFormat,
    EnumCheckOps,
    EnumEngine,
    EnumInsKey,
    EnumItm,
    EnumWriteOps,
)
from .exceptions import SPSDKUnsupportedImageType
from .images import BootImg2, BootImg3a, BootImg3b, BootImg4, BootImgRT, EnumAppType, parse
from .keystore import KeySourceType, KeyStore
from .mbi_mixin import MasterBootImageManifest, MultipleImageEntry, MultipleImageTable
from .mbimg import MasterBootImage, get_mbi_class
from .secret import MAC, CertificateImg, EnumAlgorithm, SecretKeyBlob, Signature, SrkItem, SrkTable
from .segments import (
    FlexSPIConfBlockFCB,
    PaddingFCB,
    SegAPP,
    SegBDT,
    SegCSF,
    SegDCD,
    SegIVT2,
    SegIVT3a,
    SegIVT3b,
)
from .trustzone import TrustZone, TrustZoneType

__all__ = [
    # Main Classes
    "BootImgRT",
    "BootImg2",
    "BootImg3a",
    "BootImg3b",
    "BootImg4",
    "MasterBootImage",
    # multiple images for MasterBootImage (relocation table)
    "MultipleImageEntry",
    "MultipleImageTable",
    # Segments
    "SegIVT2",
    "SegIVT3a",
    "SegIVT3b",
    "SegBDT",
    "SegAPP",
    "SegDCD",
    "SegCSF",
    "TrustZone",
    "TrustZoneType",
    "PaddingFCB",
    "FlexSPIConfBlockFCB",
    # BEE
    "BeeRegionHeader",
    "BeeKIB",
    "BeeProtectRegionBlock",
    "BeeFacRegion",
    # Secret
    "SrkTable",
    "SrkItem",
    "CertificateImg",
    "Signature",
    "MAC",
    "SecretKeyBlob",
    # Enums
    "EnumAppType",
    # Commands
    "CmdNop",
    "CmdSet",
    "CmdWriteData",
    "CmdCheckData",
    "CmdInitialize",
    "CmdInstallKey",
    "CmdAuthData",
    "CmdUnlock",
    "CmdUnlockCAAM",
    "CmdUnlockOCOTP",
    "CmdUnlockSNVS",
    # Elements
    "EnumWriteOps",
    "EnumCheckOps",
    "EnumAlgorithm",
    "EnumCertFormat",
    "EnumInsKey",
    "EnumAuthDat",
    "EnumEngine",
    "EnumItm",
    # Methods
    "parse",
    # KeyStore
    "KeyStore",
    "KeySourceType",
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Commands and responses used by MBOOT module."""

from struct import pack, unpack, unpack_from
from typing import Dict, List, Optional, Type

from spsdk.utils.easy_enum import Enum

from .error_codes import StatusCode
from .exceptions import McuBootError

########################################################################################################################
# McuBoot Commands and Responses Tags
########################################################################################################################

# fmt: off
class CommandTag(Enum):
    """McuBoot Commands."""

    NO_COMMAND                  = (0x00, "NoCommand", "No Command")
    FLASH_ERASE_ALL             = (0x01, "FlashEraseAll", "Erase Complete Flash")
    FLASH_ERASE_REGION          = (0x02, "FlashEraseRegion", "Erase Flash Region")
    READ_MEMORY                 = (0x03, "ReadMemory", "Read Memory")
    WRITE_MEMORY                = (0x04, "WriteMemory", "Write Memory")
    FILL_MEMORY                 = (0x05, "FillMemory", "Fill Memory")
    FLASH_SECURITY_DISABLE      = (0x06, "FlashSecurityDisable", "Disable Flash Security")
    GET_PROPERTY                = (0x07, "GetProperty", "Get Property")
    RECEIVE_SB_FILE             = (0x08, "ReceiveSBFile", "Receive SB File")
    EXECUTE                     = (0x09, "Execute", "Execute")
    CALL                        = (0x0A, "Call", "Call")
    RESET                       = (0x0B, "Reset", "Reset MCU")
    SET_PROPERTY                = (0x0C, "SetProperty", "Set Property")
    FLASH_ERASE_ALL_UNSECURE    = (0x0D, "FlashEraseAllUnsecure", "Erase Complete Flash and Unlock")
    FLASH_PROGRAM_ONCE          = (0x0E, "FlashProgramOnce", "Flash Program Once")
    FLASH_READ_ONCE             = (0x0F, "FlashReadOnce", "Flash Read Once")
    FLASH_READ_RESOURCE         = (0x10, "FlashReadResource", "Flash Read Resource")
    CONFIGURE_MEMORY            = (0x11, "ConfigureMemory", "Configure Quad-SPI Memory")
    RELIABLE_UPDATE             = (0x12, "ReliableUpdate", "Reliable Update")
    GENERATE_KEY_BLOB           = (0x13, "GenerateKeyBlob", "Generate Key Blob")
    FUSE_PROGRAM                = (0x14, "ProgramFuse", "Program Fuse")
    KEY_PROVISIONING            = (0x15, "KeyProvisioning", "Key Provisioning")
    TRUST_PROVISIONING          = (0x16, "TrustProvisioning", "Trust Provisioning")
    FUSE_READ                   = (0x17, "ReadFuse", "Read Fuse")
    UPDATE_LIFE_CYCLE           = (0x18, "UpdateLifeCycle", "Update Life Cycle")
    ELE_MESSAGE                 = (0x19, "EleMessage", "Send EdgeLock Enclave Message")

    # reserved commands
    CONFIGURE_I2C = (0xC1, "ConfigureI2c", "Configure I2C")
    CONFIGURE_SPI = (0xC2, "ConfigureSpi", "Configure SPI")
    CONFIGURE_CAN = (0xC3, "ConfigureCan", "Configure CAN")



class CommandFlag(Enum):
    """Flags for McuBoot commands."""

    NONE           = (0, "NoFlags", "No flags specified")
    HAS_DATA_PHASE = (1, "DataPhase", "Command has a data phase")



class ResponseTag(Enum):
    """McuBoot Responses to Commands."""

    GENERIC                     = (0xA0, "GenericResponse", "Generic Response")
    READ_MEMORY                 = (0xA3, "ReadMemoryResponse", "Read Memory Response")
    GET_PROPERTY                = (0xA7, "GetPropertyResponse", "Get Property Response")
    FLASH_READ_ONCE             = (0xAF, "FlashReadOnceResponse", "Flash Read Once Response")
    FLASH_READ_RESOURCE         = (0xB0, "FlashReadResourceResponse", "Flash Read Resource Response")
    KEY_BLOB_RESPONSE           = (0xB3, "CreateKeyBlobResponse", "Create Key Blob")
    KEY_PROVISIONING_RESPONSE   = (0xB5, "KeyProvisioningResponse", "Key Provisioning Response")
    TRUST_PROVISIONING_RESPONSE = (0xB6, "TrustProvisioningResponse", "Trust Provisioning Response")


class KeyProvOperation(Enum):
    """Type of key provisioning operation."""

    ENROLL              = (0, "Enroll", "Enroll Operation")
    SET_USER_KEY        = (1, "SetUserKey", "Set User Key Operation")
    SET_INTRINSIC_KEY   = (2, "SetIntrinsicKey", "Set Intrinsic Key Operation")
    WRITE_NON_VOLATILE  = (3, "WriteNonVolatile", "Write Non Volatile Operation")
    READ_NON_VOLATILE   = (4, "ReadNonVolatile", "Read Non Volatile Operation")
    WRITE_KEY_STORE     = (5, "WriteKeyStore", "Write Key Store Operation")
    READ_KEY_STORE      = (6, "ReadKeyStore", "Read Key Store Operation")


class KeyProvUserKeyType(Enum):
    """Enumeration of supported user keys in PUF. Keys are SoC specific, not all will be supported for the processor."""

    OTFADKEK        = (2, "OTFADKEK", "Key for OTFAD encryption")  # used on RTxxx
    SBKEK           = (3, "SBKEK", "Key for SB file encryption")  # Available on LPC55Sxx and RTxxx
    PRINCE_REGION_0 = (7, "PRINCE0", "Key for Prince region 0")  # LPC55Sxx
    PRINCE_REGION_1 = (8, "PRINCE1", "Key for Prince region 1")  # LPC55Sxx
    PRINCE_REGION_2 = (9, "PRINCE2", "Key for Prince region 2")  # LPC55Sxx
    PRINCE_REGION_3 = (10, "PRINCE3", "Key for Prince region 3")  # NHS52xx

    USERKEK         = (11, "USERKEK", "Encrypted boot image key")  # LPC55Sxx and RTxxx
    UDS             = (12, "UDS", "Universal Device Secret for DICE")  # LPC55Sxx and RTxxx


class GenerateKeyBlobSelect(Enum):
    """Key selector for the generate-key-blob function.

    For devices with SNVS, valid options of [key_sel] are
    0, 1 or OTPMK: OTPMK from FUSE or OTP(default),
    2 or ZMK: ZMK from SNVS,
    3 or CMK: CMK from SNVS,
    For devices without SNVS, this option will be ignored.
    """

    OPTMK   = (0, "OPTMK", "OTPMK from FUSE or OTP(default)")
    ZMK     = (2, "ZMK", "ZMK from SNVS")
    CMK     = (3, "CMK", "CMK from SNVS")


class TrustProvOperation(Enum):
    """Operations supported by Trust Provisioning flow."""

    PROVE_GENUINITY = (0xF4, "ProveGenuinity", "Start the proving genuinity process")
    ISP_SET_WRAPPED_DATA = (0xF0, "SetWrappedData", "Start processing Wrapped data")
    """Type of trust provisioning operation."""

    OEM_GEN_MASTER_SHARE        = (0, "OemGenMasterShare", "Enroll Operation")
    OEM_SET_MASTER_SHARE        = (1, "SetUserKey", "Set User Key Operation")
    OEM_GET_CUST_CERT_DICE_PUK  = (2, "SetIntrinsicKey", "Set Intrinsic Key Operation")
    HSM_GEN_KEY                 = (3, "HsmGenKey", "HSM gen key")
    HSM_STORE_KEY               = (4, "HsmStoreKey", "HSM store key")
    HSM_ENC_BLOCK               = (5, "HsmEncBlock", "HSM Enc block")
    HSM_ENC_SIGN                = (6, "HsnEncSign", "HSM enc sign")


class TrustProvOemKeyType(Enum):
    """Type of oem key type definition."""

    MFWISK      = (0xC3A5, "MFWISK", "ECDSA Manufacturing Firmware Signing Key")
    MFWENCK     = (0xA5C3, "MFWENCK", "CKDF Master Key for Manufacturing Firmware Encryption Key")
    GENSIGNK    = (0x5A3C, "GENSIGNK", "Generic ECDSA Signing Key")
    GETCUSTMKSK = (0x3C5A, "GETCUSTMKSK", "CKDF Master Key for Production Firmware Encryption Key")


class TrustProvKeyType(Enum):
    """Type of key type definition."""

    CKDFK = (1, "CKDFK", "CKDF Master Key")
    HKDFK = (2, "HKDFK", "HKDF Master Key")
    HMACK = (3, "HMACK", "HMAC Key")
    CMACK = (4, "CMACK", "CMAC Key")
    AESK  = (5, "AESK", "AES Key")
    KUOK  = (6, "KUOK", "Key Unwrap Only Key")


class TrustProvWrappingKeyType(Enum):
    """Type of wrapping key type definition."""

    INT_SK = (0x10, "INT_SK", "The wrapping key for wrapping of MFG_CUST_MK_SK0_BLOB")
    EXT_SK = (0x11, "EXT_SK", "The wrapping key for wrapping of MFG_CUST_MK_SK0_BLOB")

# fmt: on

########################################################################################################################
# McuBoot Command and Response packet classes
########################################################################################################################


class CmdHeader:
    """McuBoot command/response header."""

    SIZE = 4

    def __init__(self, tag: int, flags: int, reserved: int, params_count: int) -> None:
        """Initialize the Command Header.

        :param tag: Tag indicating the command, see: `CommandTag` class
        :param flags: Flags for the command
        :param reserved: Reserved?
        :param params_count: Number of parameter for the command
        """
        self.tag = tag
        self.flags = flags
        self.reserved = reserved
        self.params_count = params_count

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, CmdHeader) and vars(obj) == vars(self)

    def __ne__(self, obj: object) -> bool:
        return not self.__eq__(obj)

    def __str__(self) -> str:
        return f"<Tag=0x{self.tag:02X}, Flags=0x{self.flags:02X}, ParamsCount={self.params_count}>"

    def __repr__(self) -> str:
        return (
            f"CmdHeader(tag=0x{self.tag:02X}, flags=0x{self.flags:02X}, "
            f"reserved={self.reserved}, params_count={self.params_count})"
        )

    def to_bytes(self) -> bytes:
        """Serialize header into bytes."""
        return pack("4B", self.tag, self.flags, self.reserved, self.params_count)

    @classmethod
    def from_bytes(cls, data: bytes, offset: int = 0) -> "CmdHeader":
        """Deserialize header from bytes.

        :param data: Input data in bytes
        :param offset: The offset of input data
        :return: De-serialized CmdHeader object
        :raises McuBootError: Invalid data format
        """
        if len(data) < 4:
            raise McuBootError(f"Invalid format of RX packet (data length is {len(data)} bytes)")
        return cls(*unpack_from("4B", data, offset))


class CmdPacket:
    """McuBoot command packet format class."""

    SIZE = 32
    EMPTY_VALUE = 0x00

    def __init__(self, tag: int, flags: int, *args: int, data: Optional[bytes] = None) -> None:
        """Initialize the Command Packet object.

        :param tag: Tag identifying the command
        :param flags: Flags used by the command
        :param args: Arguments used by the command
        :param data: Additional data, defaults to None
        """
        self.header = CmdHeader(tag, flags, 0, len(args))
        self.params = list(args)
        if data is not None:
            if len(data) % 4:
                data += b"\0" * (4 - len(data) % 4)
            self.params.extend(unpack_from(f"<{len(data) // 4}I", data))
            self.header.params_count = len(self.params)

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, CmdPacket) and vars(obj) == vars(self)

    def __ne__(self, obj: object) -> bool:
        return not self.__eq__(obj)

    def __str__(self) -> str:
        return "<" + self.info() + ">"

    def info(self) -> str:
        """Get object info."""
        tag = CommandTag.get(self.header.tag, f"0x{self.header.tag:02X}")
        return f"Tag={tag}, Flags=0x{self.header.flags:02X}" + "".join(
            f", P[{n}]=0x{param:08X}" for n, param in enumerate(self.params)
        )

    def to_bytes(self, padding: bool = True) -> bytes:
        """Serialize CmdPacket into bytes.

        :param padding: If True, add padding to specific size
        :return: Serialized object into bytes
        """
        self.header.params_count = len(self.params)
        data = self.header.to_bytes()
        data += pack(f"<{self.header.params_count}I", *self.params)
        if padding and len(data) < self.SIZE:
            data += bytes([self.EMPTY_VALUE] * (self.SIZE - len(data)))
        return data


class CmdResponse:
    """McuBoot response base format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Command Response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        assert isinstance(header, CmdHeader)
        assert isinstance(raw_data, (bytes, bytearray))
        self.header = header
        self.raw_data = raw_data
        (status,) = unpack_from("<L", raw_data)
        self.status: int = status

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, CmdResponse) and vars(obj) == vars(self)

    def __ne__(self, obj: object) -> bool:
        return not self.__eq__(obj)

    def __str__(self) -> str:
        return "<" + self.info() + ">"

    def info(self) -> str:
        """Get object info."""
        return (
            f"Tag=0x{self.header.tag:02X}, Flags=0x{self.header.flags:02X}"
            + " ["
            + ", ".join(f"{b:02X}" for b in self.raw_data)
            + "]"
        )


class GenericResponse(CmdResponse):
    """McuBoot generic response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Generic response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, tag = unpack_from("<2I", raw_data)
        self.cmd_tag: int = tag

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        cmd = CommandTag.get(self.cmd_tag, f"Unknown[0x{self.cmd_tag:02X}]")
        return f"Tag={tag}, Status={status}, Cmd={cmd}"


class GetPropertyResponse(CmdResponse):
    """McuBoot get property response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Get-Property response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, *values = unpack_from(f"<{self.header.params_count}I", raw_data)
        self.values: List[int] = list(values)

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}" + "".join(
            f", v{n}=0x{value:08X}" for n, value in enumerate(self.values)
        )


class ReadMemoryResponse(CmdResponse):
    """McuBoot read memory response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Read-Memory response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, length = unpack_from("<2I", raw_data)
        self.length: int = length

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}, Length={self.length}"


class FlashReadOnceResponse(CmdResponse):
    """McuBoot flash read once response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Flash-Read-Once response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, length, *values = unpack_from(f"<{self.header.params_count}I", raw_data)
        self.length: int = length
        self.values: List[int] = list(values)
        self.data = raw_data[8 : 8 + self.length] if self.length > 0 else b""

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}, Length={self.length}"


class FlashReadResourceResponse(CmdResponse):
    """McuBoot flash read resource response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Flash-Read-Resource response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, length = unpack_from("<2I", raw_data)
        self.length: int = length

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}, Length={self.length}"


class KeyProvisioningResponse(CmdResponse):
    """McuBoot Key Provisioning response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Key-Provisioning response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, length = unpack_from("<2I", raw_data)
        self.length: int = length

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}, Length={self.length}"


class TrustProvisioningResponse(CmdResponse):
    """McuBoot Trust Provisioning response format class."""

    def __init__(self, header: CmdHeader, raw_data: bytes) -> None:
        """Initialize the Trust-Provisioning response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(header, raw_data)
        _, *values = unpack(f"<{self.header.params_count}I", raw_data)
        self.values: List[int] = list(values)

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.header.tag)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}"


class NoResponse(CmdResponse):
    """Special internal case when no response is provided by the target."""

    def __init__(self, cmd_tag: int) -> None:
        """Create a NoResponse to an command that was issued, indicated by its tag.

        :param cmd_tag: Tag of the command that preceded the no-response from target
        """
        header = CmdHeader(tag=cmd_tag, flags=0, reserved=0, params_count=0)
        raw_data = pack("<L", StatusCode.NO_RESPONSE)
        super().__init__(header, raw_data)


def parse_cmd_response(data: bytes, offset: int = 0) -> CmdResponse:
    """Parse command response.

    :param data: Input data in bytes
    :param offset: The offset of input data
    :return: De-serialized object from data
    """
    known_response: Dict[int, Type[CmdResponse]] = {
        ResponseTag.GENERIC: GenericResponse,
        ResponseTag.GET_PROPERTY: GetPropertyResponse,
        ResponseTag.READ_MEMORY: ReadMemoryResponse,
        ResponseTag.FLASH_READ_RESOURCE: FlashReadResourceResponse,
        ResponseTag.FLASH_READ_ONCE: FlashReadOnceResponse,
        ResponseTag.KEY_BLOB_RESPONSE: ReadMemoryResponse,  # not sure what format is returned, this work on RT1050
        ResponseTag.KEY_PROVISIONING_RESPONSE: KeyProvisioningResponse,
        ResponseTag.TRUST_PROVISIONING_RESPONSE: TrustProvisioningResponse,
    }
    header = CmdHeader.from_bytes(data, offset)
    if header.tag in known_response:
        return known_response[header.tag](header, data[CmdHeader.SIZE :])

    return CmdResponse(header, data[CmdHeader.SIZE :])



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for serial communication with a target device using SDP protocol."""

import logging
from typing import List, Optional, Union

from serial import Serial, SerialTimeoutException
from serial.tools.list_ports import comports

from spsdk.exceptions import SPSDKError
from spsdk.sdp.commands import CmdPacket, CmdResponse
from spsdk.sdp.exceptions import SdpConnectionError
from spsdk.utils.exceptions import SPSDKTimeoutError

from .base import SDPInterface

logger = logging.getLogger(__name__)


def scan_uart(
    port: Optional[str] = None, baudrate: Optional[int] = None, timeout: Optional[int] = None
) -> List["Uart"]:
    """Scan connected serial ports.

    Returns list of serial ports with devices that respond to PING command.
    If 'port' is specified, only that serial port is checked
    If no devices are found, return an empty list.

    :param port: name of preferred serial port, defaults to None
    :param baudrate: speed of the UART interface, defaults to 56700
    :param timeout: timeout in milliseconds, defaults to 5000
    :return: list of interfaces responding to the PING command
    """
    baudrate = baudrate or 115200
    timeout = timeout or 5000
    if port:
        interface = _check_port(port, baudrate, timeout)
        return [interface] if interface else []
    all_ports = [
        _check_port(comport.device, baudrate, timeout) for comport in comports(include_links=True)
    ]
    return list(filter(None, all_ports))


def _check_port(port: str, baudrate: int, timeout: int) -> Optional["Uart"]:
    """Check if device on comport 'port' could be opened.

    :param port: name of port to check
    :param baudrate: speed of the UART interface, defaults to 56700
    :param timeout: timeout in milliseconds
    :return: None if device can't be opened, instance of Interface if it does
    """
    try:
        logger.debug(f"Checking port: {port}, baudrate: {baudrate}, timeout: {timeout}")
        interface = Uart(port=port, baudrate=baudrate, timeout=timeout)
        return interface
    except Exception as e:  # pylint: disable=broad-except
        logger.error(str(e))
        return None


########################################################################################################################
# UART Interface Class
########################################################################################################################


class Uart(SDPInterface):
    """UART interface."""

    @property
    def is_opened(self) -> bool:
        """Return True if device is open, False otherwise."""
        return self.device.is_open

    def __init__(self, port: Optional[str] = None, timeout: int = 5000, baudrate: int = 115200):
        """Initialize the UART interface.

        :param port: name of the serial port, defaults to None
        :param baudrate: speed of the UART interface, defaults to 115200
        :param timeout: read/write timeout in milliseconds, defaults to 1000
        :raises SdpConnectionError: when there is no port available
        """
        super().__init__()
        try:
            timeout_s = timeout / 1000
            self.device = Serial(
                port=port, timeout=timeout_s, write_timeout=timeout_s, baudrate=baudrate
            )
            self.expect_status = True
        except Exception as e:
            raise SdpConnectionError(str(e)) from e

    def open(self) -> None:
        """Open the UART interface.

        :raises SdpConnectionError: when opening device fails
        """
        if not self.device.is_open:
            try:
                self.device.open()
            except Exception as e:
                raise SdpConnectionError(str(e)) from e

    def close(self) -> None:
        """Close the UART interface.

        :raises SdpConnectionError: when closing device fails
        """
        if self.device.is_open:
            try:
                self.device.close()
            except Exception as e:
                raise SdpConnectionError(str(e)) from e

    def info(self) -> str:
        """Return information about the UART interface.

        :return: information about the UART interface
        :raises SdpConnectionError: when information can not be collected from device
        """
        try:
            return self.device.port
        except Exception as e:
            raise SdpConnectionError(str(e)) from e

    def conf(self, config: dict) -> None:
        """Configure device.

        :param config: parameters dictionary
        """

    def read(self, length: Optional[int] = None) -> CmdResponse:
        """Read data from device.

        :return: data read from device
        """
        hab_info = self._read(length or 4)
        # raw_data = self._read(4) if self.expect_status else hab_info
        # return CmdResponse(hab_info, raw_data)
        return CmdResponse(self.expect_status, hab_info)

    def write(self, packet: Union[CmdPacket, bytes]) -> None:
        """Write data to the device; data might be in format of 'CmdPacket' or bytes.

        :param packet: Packet to send
        """
        self.expect_status = True
        if isinstance(packet, CmdPacket):
            data = packet.to_bytes()
        else:
            data = packet
        self._write(data)

    def _read(self, length: int) -> bytes:
        """Read 'length' amount of bytes from the device.

        :param length: Number of bytes to read
        :return: Data read
        :raises SdpConnectionError: when read from device fails
        """
        try:
            data = self.device.read(length)
            if not data:
                raise SPSDKError("No response from SDP device.")
            logger.debug(f"<{' '.join(f'{b:02x}' for b in data)}>")
            return data
        except Exception as e:
            raise SdpConnectionError(str(e)) from e

    def _write(self, data: bytes) -> None:
        """Send data to device.

        :param data: Data to send
        :raises SPSDKTimeoutError: when sending of data times-out
        :raises SdpConnectionError: when send data to device fails
        """
        logger.debug(f"[{' '.join(f'{b:02x}' for b in data)}]")
        try:
            self.device.reset_input_buffer()
            self.device.reset_output_buffer()
            self.device.write(data)
            self.device.flush()
        except SerialTimeoutException as e:
            raise SPSDKTimeoutError(
                f"Write timeout error. The timeout is set to {self.device.write_timeout} s. Consider increasing it."
            ) from e
        except Exception as e:
            raise SdpConnectionError(str(e)) from e



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Common cryptographic functions."""
import math
from typing import List, Optional

from cryptography.hazmat.primitives.asymmetric import utils
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicNumbers
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import Certificate

from spsdk import SPSDKError
from spsdk.crypto import EllipticCurvePublicKey, Encoding, PublicFormat, PublicKey
from spsdk.crypto.signature_provider import SignatureProvider
from spsdk.sbfile.misc import SecBootBlckSize

from ...exceptions import SPSDKValueError
from .abstract import BackendClass
from .backend_openssl import openssl_backend


def crypto_backend() -> BackendClass:
    """Return default crypto backend instance."""
    return openssl_backend


class Counter:
    """AES counter with specified counter byte ordering and customizable increment."""

    @property
    def value(self) -> bytes:
        """Initial vector for AES encryption."""
        return self._nonce + self._ctr.to_bytes(4, self._ctr_byteorder_encoding)  # type: ignore[arg-type]

    def __init__(
        self,
        nonce: bytes,
        ctr_value: Optional[int] = None,
        ctr_byteorder_encoding: str = "little",
    ):
        """Constructor.

        :param nonce: last four bytes are used as initial value for counter
        :param ctr_value: counter initial value; it is added to counter value retrieved from nonce
        :param ctr_byteorder_encoding: way how the counter is encoded into output value: either 'little' or 'big'
        :raises SPSDKError: When invalid byteorder is provided
        """
        assert isinstance(nonce, bytes) and len(nonce) == 16
        if ctr_byteorder_encoding not in ["little", "big"]:
            raise SPSDKError("Wrong byte order")
        self._nonce = nonce[:-4]
        self._ctr_byteorder_encoding = ctr_byteorder_encoding
        self._ctr = int.from_bytes(nonce[-4:], ctr_byteorder_encoding)  # type: ignore[arg-type]
        if ctr_value is not None:
            self._ctr += ctr_value

    def increment(self, value: int = 1) -> None:
        """Increment counter by specified value.

        :param value: to add to counter
        """
        self._ctr += value


def calc_cypher_block_count(size: int) -> int:
    """Calculate the amount if cypher blocks.

    :param size: Number of bytes for the cypher area
    :return: Number of blocks covering the cypher area
    """
    return SecBootBlckSize.to_num_blocks(size)


def matches_key_and_cert(priv_key: bytes, cert: Certificate) -> bool:
    """Verify that given private key matches the public certificate.

    :param priv_key: to be tested; decrypted binary data in PEM format
    :param cert: to be used for verification
    :return: True if yes; False otherwise
    """
    signature = crypto_backend().rsa_sign(priv_key, bytes())
    assert isinstance(cert, Certificate)
    cert_pub_key = cert.public_key()  # public key of last certificate
    assert isinstance(cert_pub_key, RSAPublicKey)
    return crypto_backend().rsa_verify(
        cert_pub_key.public_numbers().n,
        cert_pub_key.public_numbers().e,
        signature,
        bytes(),
    )


def serialize_ecc_signature(signature: bytes, coordinate_length: int) -> bytes:
    """Re-format ECC ANS.1 DER signature into the format used by ROM code."""
    r, s = utils.decode_dss_signature(signature)

    r_bytes = r.to_bytes(coordinate_length, "big")
    s_bytes = s.to_bytes(coordinate_length, "big")
    return r_bytes + s_bytes


def ecc_public_numbers_to_bytes(
    public_numbers: EllipticCurvePublicNumbers, length: Optional[int] = None
) -> bytes:
    """Converts public numbers from ECC key into bytes.

    :param public_numbers: instance of ecc public numbers
    :param length: length of bytes object to use
    :return: bytes representation
    """
    x: int = public_numbers.x
    y: int = public_numbers.y
    length = length or math.ceil(x.bit_length() / 8)
    x_bytes = x.to_bytes(length, "big")
    y_bytes = y.to_bytes(length, "big")
    return x_bytes + y_bytes


def get_matching_key_id(public_keys: List[PublicKey], signature_provider: SignatureProvider) -> int:
    """Get index of public key that match to given private key.

    :param public_keys: List of public key used to find the match for the private key.
    :param signature_provider: Signature provider used to try to match public key index.
    :raises SPSDKValueError: No match found.
    :return: Index of public key.
    """
    for i, public_key in enumerate(public_keys):
        if isinstance(public_key, RSAPublicKey):
            public_key_bytes = public_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.PKCS1
            )
        if isinstance(public_key, EllipticCurvePublicKey):
            public_key_bytes = public_key.public_bytes(
                encoding=Encoding.PEM, format=PublicFormat.SubjectPublicKeyInfo
            )

        if signature_provider.verify_public_key(public_key_bytes):
            return i

    raise SPSDKValueError("There is no match of private key in given list.")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""DK6 Prog CLI interface."""
import logging
import os
import sys
from typing import Dict, List

import click
import colorama
import hexdump
import prettytable

from spsdk import SPSDKError
from spsdk.apps.blhost_helper import progress_bar
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import spsdk_apps_common_options
from spsdk.apps.utils.utils import (
    INT,
    WARNING_MSG,
    catch_spsdk_error,
    parse_file_and_size,
    parse_hex_data,
)
from spsdk.dk6.commands import MemoryAccessValues, MemoryId, MemoryType
from spsdk.dk6.dk6device import DK6Device, DK6Memory
from spsdk.dk6.driver import Backend, DriverInterface
from spsdk.utils.misc import value_to_int

MEMORY_IDS = {
    "flash": 0,
    "psect": 1,
    "pflash": 2,
    "config": 3,
    "efuse": 4,
    "rom": 5,
    "ram0": 6,
    "ram1": 7,
}

logger = logging.getLogger(__name__)


def print_memory_table(memories: Dict[int, DK6Memory]) -> str:
    """Prints the List of Interfaces to nice colored table."""
    if len(memories) == 0:
        return (
            colorama.Fore.RED
            + "Nothing to print - empty interface list!"
            + colorama.Style.RESET_ALL
        )

    header = [
        "Memory",
        "Memory ID",
        "Base Address",
        "Length",
        "Sector Size",
        "Memory Type",
        "Access",
    ]

    table = prettytable.PrettyTable(header)
    table.align = "l"
    table.header = True
    table.border = True
    table.hrules = prettytable.HEADER
    table.vrules = prettytable.NONE
    for memory_id, memory in memories.items():
        fields = [
            colorama.Fore.YELLOW + str(MemoryId.get(memory_id)),
            colorama.Fore.MAGENTA + str(memory.mem_id),
            colorama.Fore.GREEN + hex(memory.base_address),
            colorama.Fore.RED + hex(memory.length),
            colorama.Fore.MAGENTA + hex(memory.sector_size),
            colorama.Fore.CYAN + str(MemoryType.get(memory.mem_type)),
            colorama.Fore.BLUE + str(MemoryAccessValues.get(memory.access)),
        ]

        table.add_row(fields)

    return table.get_string() + colorama.Style.RESET_ALL


def parse_memory_id(memory_id: str) -> MemoryId:
    """Convert the memory id as name or stringified number into integer.

    :param memory_id: Name or number of the property tag
    :return: memory id integer tag
    """
    try:
        value = value_to_int(memory_id)
        return MemoryId.from_int(value) if value in MEMORY_IDS.values() else MemoryId.FLASH
    except SPSDKError:
        return MemoryId.from_int(MEMORY_IDS.get(memory_id.lower(), 0))


def _split_string(string: str, length: int) -> List:
    """Split the string into chunks of same length."""
    return [string[i : i + length] for i in range(0, len(string), length)]


def format_raw_data(data: bytes, use_hexdump: bool = False, line_length: int = 16) -> str:
    """Format bytes data into human-readable form.

    :param data: Data to format
    :param use_hexdump: Use hexdump with addresses and ASCII, defaults to False
    :param line_length: bytes per line, defaults to 32
    :return: formatted string (multilined if necessary)
    """
    if use_hexdump:
        return hexdump.hexdump(data, result="return")
    data_string = data.hex()
    parts = [_split_string(line, 2) for line in _split_string(data_string, line_length * 2)]
    result = "\n".join(" ".join(line) for line in parts)
    return result


def get_dk6(ctx: click.Context) -> DK6Device:
    """Get initialized DK6 Device from click context.

    :param ctx: click Context
    :raises SPSDKError: if the device ID is not specified
    :return: DK6Device
    """
    dk6: DK6Device = ctx.obj["dk6"]
    if dk6 is None:
        raise SPSDKError("You have to specify DEVICE ID")
    dk6.init()

    return dk6


def get_default_backend() -> Backend:
    """Return default backend based on the operating system.

    :return: Backend
    """
    try:
        backend = {
            "win32": Backend.FTD2xx,
            "linux": Backend.PYFTDI,
            "darwin": Backend.PYFTDI,
        }[sys.platform]
    except KeyError:
        logger.warning(
            f"Platform {sys.platform} is not supported. Only Windows, Linux or Mac are supported."
        )
        backend = Backend.PYFTDI
    return backend


@click.group(name="dk6prog", chain=True, no_args_is_help=True)
@spsdk_apps_common_options
@click.option(
    "-d",
    "--device-id",
    help="DK6 Serial device ID, obtained by list command",
)
@click.option(
    "-r",
    "--baudrate",
    default=115200,
    help="Serial port baudrate. Default baud rate is 115200.",
)
@click.option(
    "-b",
    "--backend",
    "backend",
    type=click.Choice(Backend.__members__),  # type: ignore
    callback=lambda c, p, v: getattr(Backend, v) if v else None,
    help="PYFTDI backend, pure Python implementation of libFTDI.\n"
    + "PYLIBFTDI backend. Ctypes wrapper for libFTDI.\n"
    + "FTD2XX backend. Ctypes wrapper for D2XX.\n"
    + "PYSERIAL backend for simple UART",
)
@click.option(
    "-n",
    "--no-isp",
    "no_isp",
    is_flag=True,
    default=False,
    help="Do not send ISP sequence",
)
@click.pass_context
def main(
    ctx: click.Context,
    device_id: str,
    backend: Backend,
    baudrate: int,
    log_level: int,
    no_isp: bool,
) -> int:
    """Tool for reading and programming flash memory of DK6 target devices.

    !!! THIS IS AN EXPERIMENTAL UTILITY! USE WITH CAUTION !!!
    """
    spsdk_logger.install(level=log_level)
    click.echo(WARNING_MSG)
    backend = backend or get_default_backend()
    interface = DriverInterface(backend)
    if device_id is not None:
        if not no_isp:
            interface.go_to_isp(device_id)
        interface.init_serial(device_id, baudrate)
        dk6 = DK6Device(interface.get_serial())
        dk6.init()
    else:
        dk6 = None

    ctx.obj = {"backend": backend, "interface": interface, "dk6": dk6, "device_id": device_id}
    return 0


@main.command()
@click.pass_context
def listdev(ctx: click.Context) -> None:
    """Prints the information about the connected devices.

    DEVICE_ID identifier is a parameter used for identification of device in other commands.
    """
    interface: DriverInterface = ctx.obj["interface"]
    click.echo("List of available devices:")
    for device in interface.list_devices():
        click.echo(device)


@main.command()
@click.pass_context
def isp(ctx: click.Context) -> None:
    """Issues ISP sequence as defined in Driver interface."""
    interface: DriverInterface = ctx.obj["interface"]
    device_id: str = ctx.obj["device_id"]
    interface.go_to_isp(device_id)


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("length", type=INT(), required=True)
@click.argument("memory_id", type=str, default="0", required=False)
@click.option(
    "-o",
    "--out-file",
    metavar="FILE",
    type=click.File("wb"),
    required=False,
    help="Save output to file",
)
@click.option("-h", "--use-hexdump", is_flag=True, default=False, help="Use hexdump format")
@click.option(
    "-r",
    "--relative",
    is_flag=True,
    default=False,
    help="Use address relative to memory base address",
)
@click.pass_context
def read(
    ctx: click.Context,
    address: int,
    length: int,
    memory_id: str,
    out_file: click.File,
    use_hexdump: bool,
    relative: bool,
) -> None:
    """Reads the memory and writes it to the file or stdout.

    Returns the contents of memory at the given <ADDRESS>, for a specified <BYTE_COUNT>.

    \b
    ADDRESS     - starting address
    BYTE_COUNT  - number of bytes to read
    FILE        - store result into this file, if not specified use stdout.
                 If needed specify MEMORY_ID argument, use '-' instead name of file
                 to print to stdout.
    MEMORY_ID   - id of memory to read from (default: 0)

    \b
    Available Memory IDs:
     0 or 'PFLASH'
     1 or 'pSECT'
     2 or 'pFlash'
     3 or 'Config'
     4 or 'EFUSE'
     5 or 'ROM'
     6 or 'RAM0'
     7 or 'RAM1'

    """
    dk6 = get_dk6(ctx)

    memory = parse_memory_id(memory_id)

    with progress_bar(label="Reading memory") as progress_callback:
        data = dk6.read_memory(
            memory,
            address,
            length,
            progress_callback=progress_callback,
            relative=relative,
        )

    click.echo(
        f"Read {len(data)}/{length} bytes from {hex(address)}:{hex(address+len(data))} Memory ID: {memory_id}"
    )
    if len(data) > 0:
        if out_file and out_file.name != "<stdout>":
            click.echo(f"Writing data to {out_file.name}")
            out_file.write(data)  # type: ignore
        else:
            click.echo(format_raw_data(data, use_hexdump=use_hexdump))


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("data_source", metavar="FILE[,BYTE_COUNT] | {{HEX-DATA}}", type=str, required=True)
@click.argument("length", type=INT(), required=False)
@click.argument("memory_id", type=str, default="0", required=False)
@click.option(
    "-r",
    "--relative",
    is_flag=True,
    default=False,
    help="Use address relative to memory base address",
)
@click.pass_context
def write(
    ctx: click.Context,
    address: int,
    data_source: str,
    length: int,
    memory_id: str,
    relative: bool,
) -> None:
    """Write the memory.

    Writes the data to memory.

    \b
    ADDRESS     - starting address
    FILE        - write the content of this file
    LENGTH      - if specified, load only first LENGTH number of bytes from file
    HEX-DATA    - string of hex values: {{112233}}, {{11 22 33}}
    MEMORY_ID   - id of memory to read from (default: 0)

    \b
    Available Memory IDs:
     0 or 'PFLASH'
     1 or 'pSECT'
     2 or 'pFlash'
     3 or 'Config'
     4 or 'EFUSE'
     5 or 'ROM'
     6 or 'RAM0'
     7 or 'RAM1'

    """
    try:
        data = parse_hex_data(data_source)
    except SPSDKError:
        file_path, length = parse_file_and_size(data_source)
        with open(file_path, "rb") as f:
            data = f.read(length)

    if length == -1:
        length = os.stat(file_path).st_size

    dk6 = get_dk6(ctx)

    memory = parse_memory_id(memory_id)

    with progress_bar(label="Writing memory") as progress_callback:
        dk6.write_memory(
            memory,
            address,
            length,
            data,
            progress_callback=progress_callback,
            relative=relative,
        )

    click.echo(f"Writen {length} bytes to memory ID {memory_id} at address {hex(address)}")

    dk6.reset()


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("length", type=INT(), required=True)
@click.argument("memory_id", type=str, default="0", required=False)
@click.option(
    "-r",
    "--relative",
    is_flag=True,
    default=False,
    help="Use address relative to memory base address",
)
@click.option(
    "-v",
    "--verify",
    is_flag=True,
    default=False,
    help="Verify that the data were erase using blank check command",
)
@click.pass_context
def erase(
    ctx: click.Context,
    address: int,
    length: int,
    memory_id: str,
    relative: bool,
    verify: bool,
) -> None:
    """Erase the memory.

    Erase the content of memory at the given <ADDRESS>, for a specified <LENGTH>.

    \b
    ADDRESS     - starting address
    LENGTH      - count of bytes to be erased
    MEMORY_ID   - id of memory to erase (default: 0)

    \b
    Available Memory IDs:
     0 or 'PFLASH'
     1 or 'pSECT'
     2 or 'pFlash'
     3 or 'Config'
     4 or 'EFUSE'
     5 or 'ROM'
     6 or 'RAM0'
     7 or 'RAM1'

    """
    dk6 = get_dk6(ctx)

    memory = parse_memory_id(memory_id)

    with progress_bar(label="Erasing memory") as progress_callback:
        dk6.erase_memory(
            memory,
            address,
            length,
            relative=relative,
            verify=verify,
            progress_callback=progress_callback,
        )


@main.command()
@click.pass_context
def info(ctx: click.Context) -> None:
    """Prints the information about the connected device.

    :param ctx: click Context
    :raises SPSDKError: When the DEVICE ID is not specified
    """
    dk6: DK6Device = ctx.obj["dk6"]
    if dk6 is None:
        raise SPSDKError("You have to specify DEVICE ID")

    if dk6.chip_id:
        click.echo(
            f"Chip ID: {hex(dk6.chip_id.chip_id)}\nROM Version: {hex(dk6.chip_id.chip_version)}"
        )

    click.echo(f"MAC Address: {dk6.get_mac_str()}\n")
    click.echo(f"Detected DEVICE: {dk6.dev_type}\n")
    click.echo(print_memory_table(dk6.memories))


@catch_spsdk_error
def safe_main() -> None:
    """Calls the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing the SDP communication protocol."""
import logging
import math
from typing import Mapping, Optional, Tuple

from .commands import CmdPacket, CommandTag, ResponseValue
from .error_codes import StatusCode
from .exceptions import SdpCommandError, SdpConnectionError, SdpError
from .interfaces import SDPInterface

logger = logging.getLogger(__name__)


########################################################################################################################
# Serial Downloader Protocol (SDP) Class
########################################################################################################################
class SDP:
    """Serial Downloader Protocol."""

    @property
    def status_code(self) -> StatusCode:
        """Get status code from SDP."""
        return self._status_code

    @property
    def hab_status(self) -> int:
        """Get the response value from hab."""
        return self._hab_status

    @property
    def cmd_status(self) -> int:
        """Get the response value from the command."""
        return self._cmd_status

    @property
    def is_opened(self) -> bool:
        """Indicates whether the underlying interface is open.

        :return: True if device is open, False if it's closed
        """
        return self._device.is_opened

    def __init__(self, device: SDPInterface, cmd_exception: bool = False) -> None:
        """Initialize the SDP object.

        :param device: Interface to a device
        :param cmd_exception: True if commands should raise in exception, defaults to False
        """
        self._hab_status = 0
        self._cmd_exception = cmd_exception
        self._status_code = StatusCode.SUCCESS
        self._cmd_status = 0
        self._device = device

    def __enter__(self) -> "SDP":
        self.open()
        return self

    def __exit__(self, *args: Tuple, **kwargs: Mapping) -> None:
        self.close()

    def open(self) -> None:
        """Connect to i.MX device."""
        if not self._device.is_opened:
            logger.info(f"Connect: {self._device.info()}")
            self._device.open()

    def close(self) -> None:
        """Disconnect i.MX device."""
        self._device.close()

    def _process_cmd(self, cmd_packet: CmdPacket) -> bool:
        """Process Command Packet.

        :param cmd_packet: Command packet object
        :return: True if success else False.
        :raises SdpCommandError: If command failed and the 'cmd_exception' is set to True
        :raises SdpConnectionError: Timeout or Connection error
        """
        if not self.is_opened:
            logger.info("RX-CMD: Device Disconnected")
            raise SdpConnectionError("Device Disconnected !")

        logger.debug(f"TX-PACKET: {cmd_packet.info()}")
        self._status_code = StatusCode.SUCCESS

        try:
            self._device.write(cmd_packet)
            response = self._device.read()
        except Exception as exc:
            logger.debug(exc)
            logger.info("RX-CMD: Timeout Error")
            raise SdpConnectionError("Timeout Error") from exc

        logger.info(f"RX-PACKET: {response.info()}")
        if response.hab:
            self._hab_status = response.value
            if response.value != ResponseValue.UNLOCKED:
                self._status_code = StatusCode.HAB_IS_LOCKED

        return True

    def _read_status(self) -> int:
        """Read status value.

        :return: Status code
        :raises SdpConnectionError: Timeout
        """
        try:
            response = self._device.read()
            logger.info(f"RX-PACKET: {response.info()}")
        except Exception as exc:
            logger.info("RX-CMD: Timeout Error")
            raise SdpConnectionError("Timeout Error") from exc

        return response.value

    def _read_data(self, length: int) -> Optional[bytes]:
        """Read data from device.

        :param length: Count of bytes
        :return: bytes read if the read operation is successful else None
        :raises SdpCommandError: If command failed and the 'cmd_exception' is set to True
        :raises SdpConnectionError: Timeout or Connection error
        """
        max_length = 64
        data = b""
        remaining = length - len(data)
        while remaining > 0:
            try:
                self._device.expect_status = False
                response = self._device.read(min(remaining, max_length))
            except Exception as exc:
                logger.info("RX-CMD: Timeout Error")
                raise SdpConnectionError("Timeout Error") from exc

            if not response.hab:
                data += response.raw_data
            else:
                logger.debug(f"RX-DATA: {response.info()}")
                self._hab_status = response.value
                if response.value == ResponseValue.LOCKED:
                    self._status_code = StatusCode.HAB_IS_LOCKED
            remaining = length - len(data)
        return data[:length] if len(data) > length else data

    def _send_data(self, cmd_packet: CmdPacket, data: bytes) -> bool:
        """Send data to target.

        :param cmd_packet: Command packet object
        :param data: array with data to send
        :return: True if the write operation is successful
        :raises SdpCommandError: If command failed and the 'cmd_exception' is set to True
        :raises SdpConnectionError: Timeout or Connection error
        """
        if not self.is_opened:
            logger.info("TX-DATA: Device Disconnected")
            raise SdpConnectionError("Device Disconnected !")

        logger.debug(f"TX-PACKET: {cmd_packet.info()}")
        self._status_code = StatusCode.SUCCESS
        ret_val = True

        try:
            # Send Command
            self._device.write(cmd_packet)

            # Send Data
            self._device.write(data)

            # Read HAB state (locked / unlocked)
            hab_response = self._device.read()
            logger.debug(f"RX-DATA: {hab_response.info()}")
            self._hab_status = hab_response.value
            if hab_response.value != ResponseValue.UNLOCKED:
                self._hab_status = StatusCode.HAB_IS_LOCKED

            # Read Command Status
            cmd_response = self._device.read()
            logger.debug(f"RX-DATA: {cmd_response.info()}")
            self._cmd_status = cmd_response.value
            if (
                cmd_packet.tag == CommandTag.WRITE_DCD
                and cmd_response.value != ResponseValue.WRITE_DATA_OK
            ):
                self._status_code = StatusCode.WRITE_DCD_FAILURE
                ret_val = False
            elif (
                cmd_packet.tag == CommandTag.WRITE_CSF
                and cmd_response.value != ResponseValue.WRITE_DATA_OK
            ):
                self._status_code = StatusCode.WRITE_CSF_FAILURE
                ret_val = False
            elif (
                cmd_packet.tag == CommandTag.WRITE_FILE
                and cmd_response.value != ResponseValue.WRITE_FILE_OK
            ):
                self._status_code = StatusCode.WRITE_IMAGE_FAILURE
                ret_val = False

        except Exception as exc:
            logger.info("RX-CMD: Timeout Error")
            raise SdpConnectionError(str(exc)) from exc

        if not ret_val and self._cmd_exception:
            raise SdpCommandError("SendData", self.status_code)

        return ret_val

    def read(self, address: int, length: int, data_format: int = 32) -> Optional[bytes]:
        """Read value from reg/mem at specified address.

        :param address: Start address of first register
        :param length: Count of bytes
        :param data_format: Register access format 8, 16, 32 bits
        :return: Return bytes if success else None.
        """
        logger.info(f"TX-CMD: Read(address=0x{address:08X}, length={length}, format={data_format})")
        cmd_packet = CmdPacket(CommandTag.READ_REGISTER, address, data_format, length)
        if self._process_cmd(cmd_packet):
            return self._read_data(length)
        return None

    def read_safe(
        self,
        address: int,
        length: Optional[int] = None,
        data_format: int = 32,
        align_count: bool = False,
    ) -> Optional[bytes]:
        """Read value from reg/mem at specified address.

        This method is safe, because is validating input arguments and prevents fault execution.

        :param address: Start address of first register
        :param length: Count of bytes
        :param data_format: Register access format 8, 16, 32 bits
        :param align_count: Align the count to data_format , default False
        :return: Return bytes if success else None.
        :raises SdpError: If the address is not properly aligned
        """
        if data_format not in [8, 16, 32]:
            raise SdpError(f"Invalid data format '{data_format}'. Valid options are 8, 16, 32")
        # Check if start address value is aligned
        if (address % (data_format // 8)) > 0:
            raise SdpError(f"Address 0x{address:08X} not aligned to {data_format} bits")

        # if length is not specified, use byte-size of data_format
        length = length or data_format // 8

        # Align length value if requested
        if align_count:
            byte_alignment = data_format // 8
            length = math.ceil(length / byte_alignment) * byte_alignment

        return self.read(address, length, data_format)

    def write(self, address: int, value: int, count: int = 4, data_format: int = 32) -> bool:
        """Write value into reg/mem at specified address.

        :param address: Start address of first register
        :param value: Register value
        :param count: Count of bytes (max 4)
        :param data_format: Register access format 8, 16, 32 bits
        :return: Return True if success else False.
        :raises SdpCommandError: If command failed and the 'cmd_exception' is set to True
        """
        logger.info(
            f"TX-CMD: Write(address=0x{address:08X}, value=0x{value:08X}, count={count}, format={data_format})"
        )
        cmd_packet = CmdPacket(CommandTag.WRITE_REGISTER, address, data_format, count, value)
        if not self._process_cmd(cmd_packet):
            return False
        status = self._read_status()
        if status != ResponseValue.WRITE_DATA_OK:
            self._status_code = StatusCode.WRITE_REGISTER_FAILURE
            if self._cmd_exception:
                raise SdpCommandError("WriteRegister", self.status_code)
            return False
        return True

    def write_safe(self, address: int, value: int, count: int = 4, data_format: int = 32) -> bool:
        """Write value into reg/mem at specified address.

        This method is safe, because is validating input arguments and prevents fault execution.

        :param address: Start address of first register
        :param value: Register value
        :param count: Count of bytes (max 4)
        :param data_format: Register access format 8, 16, 32 bits
        :return: Return True if success else False.
        :raises SdpError: If the address is not properly aligned or invalid data_format
        """
        if data_format not in [8, 16, 32]:
            raise SdpError(f"Invalid data format '{data_format}'. Valid options are 8, 16, 32")
        # Check if start address value is aligned
        if (address % (data_format // 8)) > 0:
            raise SdpError(f"Address 0x{address:08X} not aligned to {data_format} bits")
        # Align count value if doesn't
        align = count % (data_format // 8)
        if align > 0:
            count += (data_format // 8) - align
        count = min(count, 4)

        return self.write(address, value, count, data_format)

    def write_csf(self, address: int, data: bytes) -> bool:
        """Write CSF Data at specified address.

        :param address: Start Address
        :param data: The CSF data in binary format
        :return: Return True if success else False.
        """
        logger.info(f"TX-CMD: WriteCSF(address=0x{address:08X}, length={len(data)})")
        cmd_packet = CmdPacket(CommandTag.WRITE_CSF, address, 0, len(data))
        return self._send_data(cmd_packet, data)

    def write_dcd(self, address: int, data: bytes) -> bool:
        """Write DCD values at specified address.

        :param address: Start Address
        :param data: The DCD data in binary format
        :return: Return True if success else False.
        """
        logger.info(f"TX-CMD: WriteDCD(address=0x{address:08X}, length={len(data)})")
        cmd_packet = CmdPacket(CommandTag.WRITE_DCD, address, 0, len(data))
        return self._send_data(cmd_packet, data)

    def write_file(self, address: int, data: bytes) -> bool:
        """Write File/Data at specified address.

        :param address: Start Address
        :param data: The boot image data in binary format
        :return: Return True if success else False.
        """
        logger.info(f"TX-CMD: WriteFile(address=0x{address:08X}, length={len(data)})")
        cmd_packet = CmdPacket(CommandTag.WRITE_FILE, address, 0, len(data))
        return self._send_data(cmd_packet, data)

    def skip_dcd(self) -> bool:
        """Skip DCD blob from loaded file.

        :return: Return True if success else False.
        :raises SdpCommandError: If command failed and the 'cmd_exception' is set to True
        """
        logger.info("TX-CMD: Skip DCD")
        cmd_packet = CmdPacket(CommandTag.SKIP_DCD_HEADER, 0, 0, 0)
        if not self._process_cmd(cmd_packet):
            return False
        status = self._read_status()
        if status != ResponseValue.SKIP_DCD_HEADER_OK:
            self._status_code = StatusCode.SKIP_DCD_HEADER_FAILURE
            if self._cmd_exception:
                raise SdpCommandError("SkipDcdHeader", self.status_code)
            return False
        return True

    def jump_and_run(self, address: int) -> bool:
        """Jump to specified address and run code from there.

        :param address: Destination address
        :return: Return True if success else False.
        """
        logger.info(f"TX-CMD: Jump To Address: 0x{address:08X}")
        cmd_packet = CmdPacket(CommandTag.JUMP_ADDRESS, address, 0, 0)
        return self._process_cmd(cmd_packet)

    def read_status(self) -> Optional[int]:
        """Read Error Status.

        :return: Return status value if success else None
        """
        logger.info("TX-CMD: ReadStatus")
        if self._process_cmd(CmdPacket(CommandTag.ERROR_STATUS, 0, 0, 0)):
            return self._read_status()
        return None

    def set_baudrate(self, baudrate: int) -> bool:
        """Configure the UART baudrate on the device side.

        The default baudrate is 115200.

        :param baudrate: Baudrate to be set
        :return: Return True if success else False.
        """
        logger.info(f"TX-CMD: Set baudrate to: {baudrate}")
        cmd_packet = CmdPacket(CommandTag.SET_BAUDRATE, baudrate, 0, 0)
        return self._process_cmd(cmd_packet)
